#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from importlib.metadata import version
try:
    VERSION_NUM = version("craterstats")
except:
    VERSION_NUM = 'non-distro version'

import numpy as np
import matplotlib.path as mpath
import craterstats.gm as gm

import importlib.resources as importlib_resources

with importlib_resources.path("craterstats.config", "demo_commands.txt") as path:
    PATH = gm.filename(str(path),'u')

PRESENTATIONS=('cumulative', 'differential', 'R-plot', 'Hartmann', 'chronology', 'rate', 'sequence', 'uncertainty')
OPLOT_TYPES=('data','differential-fit','cumulative-fit','poisson','buffered-poisson')
OPLOT_TYPES_SHORT=('data','d-fit','c-fit','poisson','b-poisson')
CRATERPLOT_KEYS=('source','name','range','snap','type','error_bars','hide','colour','psym','binning','age_left','display_age',
                 'resurf','resurf_showall','isochron','offset_age')

DEFAULT_XRANGE0 = ((-3, 2), (-3, 2), (-2, 3), (-3.15, 2.56), (4.5, 0), (4.5, 0), (4.2, 1e-2), (4.2, 1e-5))
DEFAULT_YRANGE0 = ((-4, 1), (-5, 5), (-4, 1), (-8, 5), (-6, 2), (-6, 3), (0,0),(0,0))
DEFAULT_XRANGE = dict(zip(PRESENTATIONS,DEFAULT_XRANGE0))
DEFAULT_YRANGE = dict(zip(PRESENTATIONS,DEFAULT_YRANGE0))

GREYS = (['#aaaaaa', '#e0e0e0', '#ededed', '#f8f8f8'], #standard/inverted grey values
         ['#787878', '#464646', '#373737', '#282828'])

PALETTE = ( #1st column for white background; 2nd for inverted
            ('#000000', '#ffffff', 'Black'),
            ('#a00000', '#ff0000', 'Red'),
            ('#008200', '#00ff00', 'Green'),
            ('#0050c8', '#5ab4ff', 'Blue'),
            ('#e1af00', '#ffdc00', 'Yellow'),
            ('#8748bd', '#b464ff', 'Violet'),
            ('#aaaaaa', '#787878', 'Grey'),
            ('#19426b', '#19426b', 'blue1'),
            ('#476788', '#476788', 'blue2'),
            ('#7a91a9', '#7a91a9', 'blue3'),
            ('#a0ccd3', '#a0ccd3', 'blue4'),
            ('#533714', '#533714', 'brown1'),
            ('#6b4719', '#6b4719', 'brown2'),
            ('#7b5c33', '#7b5c33', 'brown3'),
            ('#af9c83', '#af9c83', 'brown4'),
            ('#336b19', '#336b19', 'green1'),
            ('#5d8a49', '#5d8a49', 'green2'),
            ('#91af83', '#91af83', 'green3'),
            ('#b58000', '#b58000', 'orange'),
            ('#6b1946', '#6b1946', 'pink1'),
            ('#975e7d', '#975e7d', 'pink2'),
            ('#ba93a8', '#ba93a8', 'pink3'),
            ('#4c265d', '#4c265d', 'purple1'),
            ('#876e93', '#876e93', 'purple2'),
            ('#6b1920', '#6b1920', 'red1'),
            ('#7c393f', '#7c393f', 'red2'),
            ('#8d595e', '#8d595e', 'red3'),
            ('#196b64', '#196b64', 'teal1'),
            ('#6ba09b', '#6ba09b', 'teal2'),
            ('#b5a200', '#b5a200', 'yellow1'),
            ('#ded68e', '#ded68e', 'yellow2'),
            ('#a8b100', '#a8b100', 'yellow-green'),
            )

star4 = mpath.Path(
    np.array(([-1, -.25, 0, .25, 1, .25, 0, -.25, -1], [0, .25, 1, .25, 0, -.25, -1, -.25, 0])).transpose())

MARKERS = (('s', 'square', {'marker': 's', 'fillstyle': 'none', 'markersize': 3.}),
           ('o', 'circle', {'marker': 'o', 'fillstyle': 'none', 'markersize': 3.}),
           ('*4', 'star4', {'marker': star4, 'fillstyle': 'none', 'markersize': 4.3}),
           ('^', 'triangle', {'marker': '^', 'fillstyle': 'none', 'markersize': 3.5}),
           ('*', 'star5', {'marker': '*', 'fillstyle': 'none', 'markersize': 4.5}),
           ('x', 'diagonal cross', {'marker': 'x', 'markersize': 3.}),
           ('+', 'cross', {'marker': '+', 'markersize': 3.5}),
           ('.', 'point', {'marker': '.', 'markersize': 3.}),
           ('v', 'inverted triangle', {'marker': 'v', 'fillstyle': 'none', 'markersize': 3.5}),
           ('fs', 'filled square', {'marker': 's', 'markersize': 2.5}),
           ('fo', 'filled circle', {'marker': 'o', 'markersize': 2.5}),
           ('f*4', 'filled star4', {'marker': star4, 'markersize': 3.8}),
           ('ft', 'filled triangle', {'marker': '^', 'markersize': 2.5}),
           ('f*', 'filled star5', {'marker': '*', 'markersize': 3.5}),
           ('fv', 'filled inverted triangle', {'marker': 'v', 'markersize': 2.5}),
           )

ABOUT=["",gm.bright("CraterstatsIII"),
       "Copyright (c) 2021-2023, Greg Michael",
       "Version:"+VERSION_NUM,
       "",
       "Explanations of concepts and calculations used in the software are given in publications below:",
       "",
       gm.bright("Standardisation of crater count data presentation"),
       "Arvidson R.E., Boyce J., Chapman C., Cintala M., Fulchignoni M., Moore H., Neukum G., Schultz P., Soderblom L., Strom R., Woronow A., Young R. "
       "Standard techniques for presentation and analysis of crater size–frequency data. Icarus 37, 1979.",
       "",
       gm.bright("Formulation of a planetary surface chronology model"),
       "Neukum G., Meteorite bombardment and dating of planetary surfaces (English translation, 1984). Meteoritenbombardement und Datierung planetarer Oberflächen (German original, 1983).",
       "",
       gm.bright("Resurfacing correction for cumulative fits; production function differential forms"),
       "Michael G.G., Neukum G., Planetary surface dating from crater size-frequency distribution measurements: Partial resurfacing events and statistical age uncertainty. EPSL 294, 2010.",
       "",
       gm.bright("Differential fitting; binning bias correction; revised Mars epoch boundaries"),
       "Michael G.G., Planetary surface dating from crater size-frequency distribution measurements: Multiple resurfacing episodes and differential isochron fitting. Icarus, 2013.",
       "",
       gm.bright("Poisson timing analysis; mu-notation"),
       "Michael G.G., Kneissl T., Neesemann A., Planetary surface dating from crater size-frequency distribution measurements: Poisson timing analysis. Icarus, 2016.",
       "",
       gm.bright("Poisson calculation for buffered crater count"),
       "Michael G.G., Yue Z., Gou S., Di K., Dating individual several-km lunar impact craters from the rim annulus in "
       "region of planned Chang’E-5 landing Poisson age-likelihood calculation for a buffered crater counting area. EPSL 568, 2021.",
       "",
       "Full references for specific chronology or other functions are listed with the function definitions in `config/functions.txt`.",
       "",
       ]





#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

def value_locate(x,v):
    '''
    safe replacement for IDL value_locate function

    :param x:
    :param v:
    :return:
    '''
    return np.searchsorted(x,v)-1


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import matplotlib as mpl

def mpl_check_font(fontnamelist):
    '''
    matplotlib: return first available font from ordered preference list

    :param fontnamelist: e.g. ['Myriad Pro','Verdana','DejaVu Sans','Tahoma']
    :return: name of first available
    '''
    prop = mpl.font_manager.FontProperties(family=fontnamelist)
    fontfile= mpl.font_manager.findfont(prop)
    name=mpl.font_manager.get_font(fontfile).family_name
    return name


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np


def scl(x,in_range=None,out_range=None,tname=None,double=False):
    '''
    scale to integer range - output type taken from out_range if present, otherwise from x

    :param x: source array
    :param in_range: range of input data to be stretched (optional)
    :param out_range: desired output range (optional)
    :param tname: desired output type, 'byte','int16','uint16' (optional)
    :param double: if set, do calculation in doubles
    :return: rescaled data
    '''

    if tname==None:
        dtype=type(x[0]) if out_range==None else type(out_range[0])
    
    mn=min(x)
    mx=max(x)
    if in_range != None:
        mn=in_range[0]
        mx=in_range[1]
    
    if out_range == None:
      eps16 = 1e-2
      eps8 = 1e-5
      if tname == 'uint16':
          dtype=np.uint16
          out_range=[0.,65536-eps16]
      elif tname == 'int16':
          dtype = np.int16
          out_range=[-32768.,32768-eps16]
      elif tname == 'byte':
          dtype=np.ubyte
          out_range=[0.,256-eps8]
      else:
          out_range=[0.,100-eps8]

    if double:
          y=(np.float64(x)-mn)/(mx-mn)
    else:          
          y=(np.float32(x)-mn)/(mx-mn)
    
    y.clip(min=0., max=1.,out=y)
    out=out_range[0]+y*(out_range[1]-out_range[0])
     
    return np.array(out,dtype=dtype)



if __name__ == '__main__':
    a=np.linspace(2,7,4)
    print(scl(a, tname='byte'))
    print(scl(a, tname='int16'))
    print(scl(a, tname='uint16'))

    print(scl(a,out_range=[10,20]))


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np
import pandas as pd
import itertools as it
import re
import sys

import craterstats as cst
import craterstats.gm as gm

class Cratercount:
    '''Reads crater count data; provides onward in various styles and binnings'''

    BINNINGS = ['pseudo-log', '20/decade', '10/decade', 'x2', 'root-2', '4th root-2', 'none'] #allowed binnings

    def __init__(self,filename=None):
        self.filename=filename
        filetype = gm.filename(filename, 'e', max_ext_length=6) if filename else None

        self.binning=None                                                                 #current binning
        self.binned={}
        self.perimeter=None
        self.buffered=False
        self.prebinned=False

        try:
            if filetype=='.stat': self.ReadStatFile()
            elif filetype=='.diam': self.ReadDiamFile()
            # '.binned' disallowed for now
            # requires dedicated code to interact with 'range' specification beyond given bins
            # needs further consideration (e.g. for poisson range), but better not to use data in this format anyway
            #elif filetype=='.binned': self.ReadBinnedFile()
            elif filetype=='.scc': self.ReadSCCfile()
            #elif filetype=='csv': self.read_JMARS_file  #need new model data file
            elif filetype==None: pass #just create empty object
        except:
            print("Unable to read file: "+filename)
            raise SystemExit(0)

    def __str__(self):
        return self.filename

# ;****************************************************
# ;             Generate missing details
# ;****************************************************

    def MakeBinGeometricMean(self):
        d_min=self.binned['d_min']
        d_max=np.append(d_min[1:],d_min[-1]*( (d_min[-1]/d_min[-2])))

        self.binned.update({'d_mean':np.sqrt(d_min*d_max),
                            'bin_width':d_max-d_min,
                            'd_max':d_max})

    def MakeBinMin(self):
        d_mean=self.binned['d_mean']
        d_min0=[np.sqrt(a*b) for a,b in zip(d_mean,d_mean[1:])]
        d_min=[d_min0[0]*d_mean[0]/d_mean[1]]+d_min0    #assume first bin in same ratio
        self.binned.update(d_min=d_min)

        width0=[b-a for a,b in zip(d_mean,d_mean[1:])]
        width=width0+[width0[-1]*d_mean[-1]/d_mean[-2]] #assume last bin in same ratio
        self.binned.update(bin_width=width)

    def MakeBinNcum(self):
        self.binned.update(ncum=np.array(list(it.accumulate(reversed(self.binned['n'])))))
        self.binned.update(ncum_event=np.array(list(it.accumulate(reversed(self.binned['n_event'])))))


# ;****************************************************
# ;                    File readers
# ;****************************************************

    def ReadStatFile(self):

        #valid stat file has either 7 or 11 columns:
        col=['diam','n','n_density','n_density_err','ncum','ncum_density','ncum_density_err',
             'd_mean','n_diff','n_diff_err','n_event']

        t=pd.read_csv(self.filename,sep='\s+',comment='#',skip_blank_lines=True,header=None)

        nc=t.shape[1]
        t.columns=col[:t.shape[1]]

        self.area=t['ncum'][0]/t['ncum_density'][0]
        self.binned={'d_min':t['diam'].to_numpy(),
                     'n':t['n'].to_numpy(),
                     'n_event':t['n'].to_numpy() if nc==7 else t['n_event'].to_numpy(),
                     'ncum':t['ncum'].to_numpy(),
                     'ncum_event':t['ncum'].to_numpy()
                     }

        self.prebinned=True
        self.binning='unknown'
        self.MakeBinGeometricMean()


    def ReadDiamFile(self):
        s= gm.read_textstructure(self.filename)

        c=s['crater']
        diam=[float(e) for e in c['diameter']]

        if 'reference_area' in c:                           #buffered case
            area=1. #nominal area
            self.buffered=True
            frac=[area/float(e) for e in c['reference_area']]
        else:                                               #normal case
            area=float(s['area'])
            frac=[float(e) for e in c['fraction']] if 'fraction' in c else [1.]*len(diam)

        if (min(frac)<0 or max(frac)>1) and self.buffered==0:
            self.errormsg="Crater list in "+self.filename+" has invalid crater fractions."
        if area==0:
            self.errormsg="Crater list in "+self.filename+" has undefined area."


        diam,frac=zip(*sorted(zip(diam,frac),reverse=True))

        self.area=area
        self.diam=diam
        self.fraction=frac
        self.prebinned=0


    def ReadBinnedFile(self):
        s= gm.read_textstructure(self.filename)
        t=s['table']
        diam=[float(e) for e in t['diameter']]

        q=[i for i,e in sorted(enumerate(diam),key=lambda x:x[1])]     #get sorted indices

        self.area=float(s['area'])

        d_min=[diam[e] for e in q]
        n=[float(t['frequency'][e]) for e in q]
        n_event=[float(t['event_frequency'][e]) for e in q] if 'event_frequency' in t else n
        self.binned={'d_min':np.array(d_min),'n':np.array(n),'n_event':np.array(n_event)}

        self.prebinned=1
        self.MakeBinGeometricMean()
        self.MakeBinNcum()


    def ReadSCCfile(self):
        s= gm.read_textstructure(self.filename)
        c=s['crater']
        diam=[float(e) for e in c['diam']]

        frac=[float(e) for e in c['fraction']] if 'fraction' in c else [1. for e in diam]

        q=[i for i,e in sorted(enumerate(diam),key=lambda x:x[1])]     #get sorted indices

        self.area=float(re.findall('\s*[\d\.]*',s['Total_area'])[0])
        if 'Perimeter' in s.keys():
            self.perimeter = float(s['Perimeter'])
        if 'Total_perimeter' in s.keys(): #for Thomas Heyer's OpenCraterTool
            self.perimeter = float(re.findall('\s*[\d\.]*',s['Total_perimeter'])[0])
        self.diam=[diam[e] for e in q]
        self.fraction=[frac[e] for e in q]
        self.prebinned=0


# ;****************************************************
# ;                 File writers
# ;****************************************************

    def WriteStatFile(self,filename,binning='pseudo-log'):

        if not self.prebinned:
            if self.binning != binning:
                self.apply_binning(binning)

        out=['# Craterstats3 exported stat file',\
             '#--------------------------------',\
             '#','# Source: '+self.filename,'# Binning: '+self.binning,\
             '#','# Total area = '+format(self.area,'0g'),'#',\
             '#  D_min       F(D)       N_inc       Error        C(D)       N_cum       Error      D_mean      N_diff       Error     N_event','#',\
             '#------------------------------------------------------------------------------------------------------------------------------']

        a=self.area
        b=self.binned

        for i in range(len(b['d_min'])):
            if b['n'][i]>0:
                s=(format(b['d_min'][i],'<7.5g')+
                   format(b['n'][i],'>11.5g')+
                   format(b['n'][i]/a,'>13.3E')+
                   format(b['n'][i]/a/np.sqrt(b['n'][i]),'>12.3E')+
                   format(b['ncum'][i],'>11.5g')+
                   format(b['ncum'][i]/a,'>13.3E')+
                   format(b['ncum'][i]/a/np.sqrt(b['ncum'][i]),'>12.3E')+
                   format(b['d_mean'][i],'>12.4g')+
                   format(b['n'][i]/b['bin_width'][i]/a,'>12.3E')+
                   format(b['n'][i]/b['bin_width'][i]/a/np.sqrt(b['n_event'][i]),'>12.3E')+
                   format(b['n_event'][i],'>12.5g'))
                out+=[s]

        out+=['#------------------------------------------------------------------------------------------------------------------------------']
        gm.write_textfile(filename, out)

# ;****************************************************

    def generate_bins(self,binning,d,offset=0.,expand=True):
        '''
        Generate bin boundaries to cover range of d

        :param binning: binning style string
        :param d: diameter values
        :param offset: fractional offset (set 0.5 to offset by half bin)
        :param expand: create one boundary beyond data (default True)
        :return: bin boundaries
        '''
        r10=np.log10(gm.range(d))
        eps=.001 # small vs smallest possible bin factor
        if expand:
            r10[1] += eps
        else:
            r10[0] += eps
            r10[1] -= eps

        if binning == 'none': return d
        if binning == 'pseudo-log':
            a=[1.,1.1,1.2,1.3,1.4,1.5,1.7,2.,2.5,3.,3.5,4.,4.5,5.,6.,7.,8.,9.]
            b=[10**e for e in range(int(np.floor(r10[0])),int(np.ceil(r10[1]+.1)))] # if above .9 make sure get next decade
            bins=np.outer(b,a).flatten()
        else:
            bins_per_decade = {'x2': 1. / np.log10(2.),
                               'root-2': 2./np.log10(2.),
                               '4th root-2': 4./np.log10(2.),
                               '10/decade': 10.,
                               '20/decade': 20.,
                            }[binning]

            br = np.array(gm.range(r10*bins_per_decade + offset, outer=True)) - offset
            b0 = np.arange(br[0],br[1]+1)
            bins = 10 ** (b0 / bins_per_decade)

        q=np.searchsorted(bins,10**r10)
        needed_bins = bins[np.clip(q[0]-1,0,None):q[1]+1]
        return needed_bins


    def apply_binning(self,binning,offset=0.):
        '''
        Create binned version of cratercount (internally, into self.binned structure)

        :param binning: name of binning style
        :param offset: fractional offset (set 0.5 to offset by half bin)
        :return: none
        '''
        if not binning in self.BINNINGS: raise Exception("Invalid binning")

        d=self.diam

        if binning=='none':
            bins,h = zip(*sorted(zip(d, self.fraction)))
            bins, h = np.array(bins), np.array(h)
            bin_centres = bins
            h_event=np.ones(len(h))
            width = np.zeros(len(d))

        else:
            bins=self.generate_bins(binning,d,offset=offset)
            h_event, bins = np.histogram(d, bins=bins)
            h,bins=np.histogram(d,weights=self.fraction,bins=bins)

            width=bins[1:]-bins[:-1]
            bin_centres=np.sqrt(bins[1:]*bins[:-1])

        self.binned={'d_min':bins[:-1] if binning!='none' else bins,
                     'd_max':bins[1:] if binning!='none' else bins,
                     'bin_width':width,
                     'd_mean':bin_centres,
                     'n':h,
                     'n_event':h_event,
                     'ncum':np.flip(np.cumsum(np.flip(h))),
                     'ncum_event':np.flip(np.cumsum(np.flip(h_event)))
                     }
        self.binning=binning

    def decode_range(self,range,binning,snap):
        if not self.prebinned:
            if self.binning != binning:
                self.apply_binning(binning)

        r1 = [0.,0.]
        snap1 = [snap,snap]
        q = np.where(self.binned['n'] > 0)
        for i,r in enumerate(range):
            try:
                if isinstance(r, float):
                    r1[i]=r
                elif r[0]=='b':
                    v=int(r[1:])
                    v=np.clip(v,-len(q[0]),len(q[0]))
                    if v>0:
                        v-=1
                    r1[i]=self.binned['d_mean'][q[0][v]]
                    snap1[i]=1
                else:
                    r1[i]=float(r)
            except:
                sys.exit('Range limit error: ' + r)

        if r1[1] < r1[0]: r1[1]=r1[0]

        if self.binning != 'none':
            if snap1[0] and r1[0]>0:
                r1[0] = self.binned['d_min'][np.searchsorted(self.binned['d_min'],r1[0]*(1.0000001))-1]
            if snap1[1] and r1[1] <= self.binned['d_max'][-1]-1e-7:
                r1[1] = self.binned['d_max'][np.searchsorted(self.binned['d_max'],r1[1]*(0.999999))]

        return r1


# ;****************************************************
# ;                    plot output
# ;****************************************************

    def resurf_adj(self,pf,range):
        p0=self.getplotdata("cumulative",self.binning,range=range)
        p=p0.copy()
        n=len(p['d'])
        count=0
        ncum_adj=0.
        if n > 1:
            while True:
                ncum_adj0=ncum_adj
                p['y']=p0['y']+ncum_adj
                a0=pf.fit(p)
                ncum_adj=pf.evaluate("cumulative",p['d'][n-1],a0[0])-p0['y'][n-1]
                count+=1
                if abs(ncum_adj0-ncum_adj) < .0001 or count > 20:
                    break
        return ncum_adj


    def getplotdata(self,presentation,binning,
                    range=None,
                    resurfacing=None,
                    pf=None):
        '''
        Return cratercount data in one of several presentation forms suitable for plotting

        :param presentation: data presentation name
        :param binning: binning name
        :param range: diameter range (km)
        :param resurfacing: None|0|1  - no correction|show only range of correction|show all corrected points
        :param pf: production function (for binning bias or resurfacing corrections)
        :return: dictionary of plottable values
        '''

        if not self.prebinned:
            if self.binning != binning:
                self.apply_binning(binning)

        adj=0 if resurfacing is None else self.resurf_adj(pf,range)

        q=np.where(self.binned['n']>0)

        if presentation=='cumulative':
            d=self.binned['d_min'][q]
            y=self.binned['ncum'][q]/self.area+adj
            err=y/np.sqrt(self.binned['ncum_event'][q])

        elif presentation=='incremental':
            d=self.binned['d_mean'][q]
            y=self.binned['n'][q]/self.area
            err=y/np.sqrt(self.binned['n_event'][q])

        elif presentation=='Hartmann':
            d=self.binned['d_mean'][q]
            y=self.binned['n'][q]/self.binned['bin_width'][q]/self.area*d*(np.sqrt(2.)-1.)
            err=y/np.sqrt(self.binned['n_event'][q])

        elif presentation=='differential':
            d=self.binned['d_mean'][q]
            y=self.binned['n'][q]/self.binned['bin_width'][q]/self.area
            err=y/np.sqrt(self.binned['n_event'][q])

            #correct for exponential binning bias (Michael, 2013 - Section 5)
            beta=(self.binned['d_mean'][q]/self.binned['d_min'][q])**2  #bin ratio
            if pf:
                k=np.log10(pf.evaluate("cumulative",d*np.sqrt(beta))/pf.evaluate("cumulative",d/np.sqrt(beta)))/np.log10(beta)
            else:
                k=-3. #if exact slope unknown, taking -3 (cumulative) is much better than nothing
            f=cst.bin_bias_correction(beta, k)
            y/=f

        elif presentation=='R-plot':
            d=self.binned['d_mean'][q]
            y=d**3*self.binned['n'][q]/self.binned['bin_width'][q]/self.area
            err=y/np.sqrt(self.binned['n_event'][q])

        if range:
            d_mean_q=self.binned['d_mean'][q]
            q1=np.where((range[0] < d_mean_q)  & (d_mean_q < range[1]))
            d=d[q1]
            y=y[q1]
            err=err[q1]
            q=q[0][q1]  #[0] takes first dim of tuple

            r1=range.copy()
            if resurfacing == 1: r1[0]=0 # show all corrected points
            if r1[0] == 0: r1[0]=self.binned['d_min'][0]
            if r1[1] == np.Inf: r1[1] = self.binned['d_max'][-1]
            bin_range = tuple(np.array(self.generate_bins(self.binning,r1,expand=False))[[0,-1]])  # extend selection to bin boundaries
        else:
            bin_range = self.binned['d_min'][q[0][0]], self.binned['d_max'][q[0][-1]] # data range


        n=np.sum(self.binned['n'][q])             #no of craters in range
        n_event=np.sum(self.binned['n_event'][q])

        return {'presentation':presentation,'d':d,'y':y,'err':err,'n':n,'n_event':n_event,'bin_range':bin_range}



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import re

from .strip_quotes import strip_quotes

def quoted_split(s,separator='\s'):
    '''
    split string using separator characters, leaving separators within quoted substrings

    :param s: source string
    :param separator: default is whitespace; for comma with whitespace, use '\s,'
    :return: list of substrings
    '''

    return [strip_quotes(p).replace('\\"', '"').replace("\\'", "'") \
            for p in re.findall(r'"(?:\\.|[^"])*"|\'(?:\\.|[^\'])*\'|[^'+separator+']+', s)]



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from .file_exists import file_exists
from .filename import filename
from .read_textfile import read_textfile
from .read_textstructure import read_textstructure
from .write_textfile import write_textfile


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

def range(x,magnitude=False,outer=False):
    """
    Find min, max of vector x

    :param x: vector
    :param magnitude: return magnitude of range instead?
    :param outer: return integer range containing range?
    :return: (min,max)
    """
    if len(x)==0:
        return np.nan if magnitude else np.nan,np.nan

    mn=np.min(x)
    mx=np.max(x)
    
    if outer:
        mn=np.floor(mn)
        mx=np.ceil(mx)    
    
    if magnitude:
        return mx-mn  
    
    return mn,mx


def mag(x,outer=False):
    return range(x,magnitude=True,outer=outer)


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.


def strip_quotes(s):
    '''remove matching quotes from start/end of string'''

    if len(s)>1 and s[0]==s[-1] and s[0] in ['"',"'"]:
        return s[1:-1]
    else:
        return s




#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

def normal(m,s,x,cumulative=False):
    '''
    evaluate normal pdf at point(s) x

    :param m: mean
    :param s: sd
    :param x: point(s) to evaluate
    :param cumulative: return cdf instead of pdf
    :return: pdf
    '''

    if cumulative:
        return 0.5 * (1 + np.math.erf((x - m) / (s * np.sqrt(2.))))
    else:
        return np.exp(-((x-m)**2)/(2*s**2))/(s*np.sqrt(2*np.pi))

if __name__ == '__main__':
    pass
    # x=gm_scl(findgen(300),out_range=[0.,x_max])
    # plot,x,gm_normal(m,s,x,cumulative=cumulative)



#  Copyright (c) 2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

#import craterstats.gm.string.sigfigs as sigfigs
from .sigfigs import sigfigs

def scientific_notation(v,e1=None,e2=0,sf=3,force=False, MathML=False, unit=None):
    '''
    Convert number to latex/MathML-style scientific notation string

    values >0.01 and <10000 are shown without multiplier unless force=True

    :param v: value to convert
    :param e1,e2: plus/minus values
    :param sf: number of significant figures
    :param force: force notation even for low-value exponents
    :return: formatted string
    '''

    u = ('', '')
    if unit is not None: #add more as needed
        u = {'km-2':(r'$ km$^{-2}','<msup><mtext>&#8239;km</mtext><mtext>-2</mtext></msup>'),
             'km2': (r'$ km$^{2}', '<msup><mtext>&#8239;km</mtext><mtext>2</mtext></msup>')
             }[unit]

    if -2 < np.log10(abs(v)) < 4 and not force:
        st = sigfigs(v, sf)
        exp = 0
    else:
        s0 = ("{:." + str(sf-1) + "e}").format(v)
        s = s0.split('e')
        exp = int(s[1])
        st = s[0]

    if e1 is not None:
        e1s = sigfigs((e1 - v)/10**exp, sf - 1)
        e2s = sigfigs((v - e2)/10**exp, sf - 1)
        if MathML:
            st = '<msubsup><mtext>'+st+'</mtext><mtext>&#x2212;'+e2s+'</mtext><mtext>+'+e1s+'</mtext></msubsup>'
        else:
            st += '^{+' + e1s + '}_{-' + e2s + '}'

    if MathML:
        st = ('<math xmlns="http://www.w3.org/1998/Math/MathML">'+st
              +('' if exp==0 else '<mo>&#183;</mo><msup><mtext>10</mtext><mtext>'+str(exp)+'</mtext></msup>')
              +u[1]+'</math>')
    else:
        st = '$' + st + ('' if exp==0 else'\cdot10^{'+str(exp)+'}')+u[0] + '$'

    return st

if __name__ == '__main__':
    print(scientific_notation(3.14e9,3.17e9,2.6e9))
    print(scientific_notation(3.14e9,3.17e9,2.6e9,unit='km-2'))
    print(scientific_notation(3.14e9,3.17e9,2.6e9, MathML=True))
    print(scientific_notation(3.14e9,3.17e9,2.6e9,unit='km-2', MathML=True))
    print(scientific_notation(3.14,3.17,2.6))
    print(scientific_notation(3.14,3.17,2.6,unit='km-2'))
    print(scientific_notation(3.14,3.17,2.6, MathML=True))
    print(scientific_notation(3.14,3.17,2.6,unit='km-2', MathML=True))



# def scientific_notation(n,sf=3,force=False):
#     '''
#     Convert number to latex-style scientific notation string
#
#     values >0.01 and <10000 are shown without multiplier unless force=True
#
#     :param n: number to convert
#     :param sf: number of significant figures
#     :param force: force notation even for low-value exponents
#     :return: latex-style formatted string
#     '''
#
#     if -2 < np.log10(abs(n)) < 4 and not force:
#             return sigfigs(n, sf)
#
#     s0 = ("{:." + str(sf-1) + "e}").format(n)
#     s = s0.split('e')
#     return s[0]+'×10^{'+str(int(s[1]))+'}'


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

def bin_bias_correction(beta,k):
    '''
    Correction for binning bias of log-log histogram of known slope (see Michael (2013), Eq. 2)

    :param beta: ratio between adjacent bin boundaries, e.g. sqrt(2)
    :param k: log-log slope (including minus sign, e.g. -2)
    :return: correction ratio
    '''

    g=beta**.5-beta**(-.5)
    with np.errstate(invalid='ignore'): # both vectors of conditional are evaluated:
        Fbin2F_ratio = np.where(abs(k) < 1e-5,1., (beta ** (k / 2.) - beta ** (-k / 2.)) / (g * k) )

    return Fbin2F_ratio


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import unittest
from unittest.mock import patch, mock_open

import numpy as np
import colorama

import craterstats.gm as gm

class Testgm(unittest.TestCase):

    # maths

    def test_scl(self):
        a=np.linspace(2, 7, 4)
        self.assertTrue(np.allclose(gm.scl(a, out_range=[10., 20.]), [10., 13.333333, 16.666666, 20.]))
        self.assertTrue(np.array_equal(gm.scl(a, out_range=[10, 20]), [10, 13, 16, 20]))
        self.assertTrue(np.array_equal(gm.scl(a, tname='byte'), [  0,  85, 170, 255]))
        self.assertTrue(np.array_equal(gm.scl(a, tname='int16'), [-32768, -10922, 10922, 32767]))
        self.assertTrue(np.array_equal(gm.scl(a, tname='uint16'), [0, 21845, 43690, 65535]))

    def test_range(self):
        self.assertEqual(gm.range([3, 7, 2]),(2,7))
        self.assertEqual(gm.range([3.5, -7, 2]), (-7, 3.5))

    def test_mag(self):
        self.assertEqual(gm.mag([3, 7, 2]),5)
        self.assertEqual(gm.mag([3, -7, 2]), 10)

    def test_poly(self):
        self.assertTrue(np.array_equal(gm.poly([2,3,1], [0,-1,2]), [2+0+0,2+3*(-1)+(-1)**2,2+3*2+2**2]))

    def test_poisson(self):
        lam=np.array([0.,1.,1.5])
        res=(0.,
             1.**3 * np.exp(-1.) / (3*2*1),
             1.5 ** 3 * np.exp(-1.5) / (3 * 2 * 1)
             )
        self.assertEqual(gm.poisson(3,0.),res[0])
        self.assertEqual(gm.poisson(3, 1.), res[1])
        self.assertEqual(gm.poisson(3, 1.5), res[2])
        self.assertTrue(np.array_equal(gm.poisson(3, lam), res))
        self.assertTrue(np.array_equal(gm.poisson(2, lam, cumulative=True),
                                       gm.poisson(0, lam)+gm.poisson(1, lam)+gm.poisson(2, lam) ))
        self.assertTrue(np.array_equal(gm.poisson(25, 3., threshold=23), gm.normal(3., np.sqrt(3.), 25)))
        self.assertTrue(np.array_equal(gm.poisson(25, 3., threshold=23, cumulative=True),
                                       gm.normal(3., np.sqrt(3.), 25, cumulative=True) ))

    def test_normal(self):
        self.assertGreater(gm.normal(10., 1., 10.),gm.normal(10., 1., 9.9))
        self.assertEqual(gm.normal(10.,1.,10.,cumulative=True),0.5)
        self.assertAlmostEqual(gm.normal(10., 1., 11., cumulative=True) - gm.normal(10., 1., 9., cumulative=True),
                               .683,places=3)

    # plotting

    def test_mpl_check_font(self):
        fontnamelist=['Myriad Pro','Verdana','DejaVu Sans','Tahoma']
        self.assertIn(gm.mpl_check_font(fontnamelist),fontnamelist)

    def test_ticks(self):
        self.assertTrue(np.array_equal(gm.ticks([0,4.5],3), [0., 5.]))
        self.assertTrue(np.array_equal(gm.ticks([0, .9], 3), [0. , 0.5, 1. ]))
        self.assertTrue(np.array_equal(gm.ticks([0.5, 3.5], 3), [0., 2., 4.]))

    # string

    def test_bright(self):
        self.assertEqual(gm.bright('test'), colorama.Style.BRIGHT+'test'+colorama.Style.RESET_ALL)
        self.assertTrue(gm.bright.init)

    def test_strip_quotes(self):
        self.assertEqual(gm.strip_quotes('"test"'),'test')
        self.assertEqual(gm.strip_quotes("'test'"), 'test')
        self.assertEqual(gm.strip_quotes('"'), '"')
        self.assertEqual(gm.strip_quotes(''), '')

    def test_quoted_split(self):
        self.assertEqual(gm.quoted_split('"test"  0'),['test', '0'])
        self.assertEqual(gm.quoted_split('test'), ['test'])
        self.assertEqual(gm.quoted_split(r"9684  0 Inf 'test data' 1.76e+04 'D:\tmp\test.scc'"),
                         ['9684', '0', 'Inf', 'test data', '1.76e+04', r'D:\tmp\test.scc'])
        self.assertEqual(gm.quoted_split(r"9684,  0, Inf, 'test, data', 1.76e+04 'D:\tmp\test.scc'",separator='\s,'),
                         ['9684', '0', 'Inf', 'test, data', '1.76e+04', r'D:\tmp\test.scc'])

    def test_sigfigs(self):
        self.assertEqual(gm.sigfigs(3.14159, 3),'3.14')
        self.assertEqual(gm.sigfigs(3.14159, 2), '3.1')
        self.assertEqual(gm.sigfigs(112.14159, 2), '110')
        self.assertEqual(gm.sigfigs(-.0314159, 3), '-0.0314')
        self.assertEqual(gm.sigfigs(314759, 3), '315000')
        self.assertEqual(gm.sigfigs(31.4759, 1), '30')
        self.assertEqual(gm.sigfigs([243.3, 0.3422], 2), ['240', '0.34'])

    def test_scientific_notation(self):
        self.assertEqual(gm.scientific_notation(10 ** 3 * np.pi, sf=2),'$3100$')
        self.assertEqual(gm.scientific_notation(10 ** 3 * np.pi, sf=2, force=True), '$3.1\cdot10^{3}$')
        self.assertEqual(gm.scientific_notation(10 ** -2 * np.pi, sf=2),'$0.031$')
        self.assertEqual(gm.scientific_notation(10 ** -3 * -np.pi, sf=3), '$-3.14\cdot10^{-3}$')

    def test_diameter_range(self):
        self.assertEqual(gm.diameter_range(np.array([.65, 1.2])),'0.65–1.2 km')
        self.assertEqual(gm.diameter_range([1.1, 1.7]),'1.1–1.7 km')
        self.assertEqual(gm.diameter_range([.65, .95]),'650–950 m')


    #file

    def test_filename(self):
        f=r'd:\aaa\bbb\test.txt'
        self.assertEqual(gm.filename(f, 'n'), 'test')
        self.assertEqual(gm.filename(f, 'e'), '.txt')
        self.assertEqual(gm.filename(f, 'p'), 'd:\\aaa\\bbb\\')
        self.assertEqual(gm.filename(f, 'b'), 'bbb\\')
        self.assertEqual(gm.filename(f, 'u'), 'd:\\aaa\\')
        self.assertEqual(gm.filename(f, 'pn1e', '_tag'), r'd:\aaa\bbb\test_tag.txt')

    def test_write_textfile(self):
        m = mock_open()
        f = r'd:\tmp\test.txt'
        with patch('builtins.open', m):
            gm.write_textfile(f,'test-string')
            m.assert_called_once_with(f, 'w')
            m().writelines.assert_called_once_with('test-string')

            gm.write_textfile(f, ['a','b'])
            m().writelines.assert_called_with('a\nb')

    def test_read_textfile(self):
        f = r'd:\tmp\test.txt'
        c = '1\n2\n#test\n\n5 ;comment'
        with patch('builtins.open', mock_open(read_data=c)) as m:
            self.assertEqual(gm.read_textfile(f),['1', '2', '#test', '', '5 ;comment'])
            m.assert_called_once_with(f, 'r', encoding='utf-8-sig')
            self.assertEqual(gm.read_textfile(f, as_string=True),c)
            self.assertEqual(gm.read_textfile(f, n_lines=2), ['1', '2'])
            self.assertEqual(gm.read_textfile(f, ignore_hash=True), ['1', '2', '', '5 ;comment'])
            self.assertEqual(gm.read_textfile(f, ignore_blank=True), ['1', '2', '#test', '5 ;comment'])
            self.assertEqual(gm.read_textfile(f, strip=';'), ['1', '2', '#test', '','5 '])

    def test_read_textstructure(self):
        self.assertEqual(gm.read_textstructure('scalar=10', from_string=True), {'scalar': '10'})
        self.assertEqual(gm.read_textstructure("string='this is a string.'", from_string=True), {'string': 'this is a string.'})
        self.assertEqual(gm.read_textstructure('array=[1.3,15.7,6,14]', from_string=True), {'array': ['1.3', '15.7', '6', '14']})
        self.assertEqual(gm.read_textstructure('array=[1,2,\n3,4]', from_string=True), {'array': ['1', '2', '3', '4']})
        self.assertEqual(gm.read_textstructure('pointer=*[1,2,3]', from_string=True), {'pointer': ['1', '2', '3']})
        self.assertEqual(gm.read_textstructure('boundary = {lon,lat\n-126.97	17.35\n-124.45	18.94\n}',
            from_string=True), {'boundary': {'lat': ['17.35', '18.94'], 'lon': ['-126.97', '-124.45']}})
        self.assertEqual(gm.read_textstructure('wifi={mac,name\n04f02a418c2e "ABC Wi-Fi"\n}',
            from_string=True), {'wifi': {'mac': ['04f02a418c2e'], 'name': ['ABC Wi-Fi']}})
        self.assertEqual(gm.read_textstructure('text_block=""As the sun set and the moon rose,\n'
                                               'she happened upon a splendid castle."\n"',from_string=True),
            {'text_block': '"As the sun set and the moon rose,\nshe happened upon a splendid castle."\n'})
        self.assertEqual(gm.read_textstructure('dict={\n  a=30\n  b=[2,3]\n}', from_string=True), {'dict': {'a': '30', 'b': ['2', '3']}})
        self.assertEqual(gm.read_textstructure('implied=1\nimplied=2\nimplied=3', from_string=True), {'implied': ['1', '2', '3']})


if __name__ == '__main__':
    unittest.main()



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import os
import sys
import argparse
import numpy as np
import re
import pathlib

import importlib.resources as importlib_resources

import craterstats as cst
import craterstats.gm as gm


class AppendPlotDict(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        s=' '.join(values)
        d = {}
        for kv in re.split(',(?=\w+=)',s): # only split on commas directly preceding keys
            try:
                k, v = kv.split("=")
            except:
                sys.exit('Syntax error: ' + option_string+' '+kv)
            d[k] = v
        list_of_d = getattr(namespace, self.dest)
        list_of_d=[d] if list_of_d is None else list_of_d+[d]
        setattr(namespace, self.dest, list_of_d)

class SpacedString(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, ' '.join(values))

class LoadFromFile(argparse.Action):
    def __call__ (self, parser, namespace, values, option_string = None):
        with values as f:
            # parse arguments in the file and store them in the target namespace
            a=f.read()
            a=('\n').join([e for e in a.split('\n') if e.strip() and (e+' ')[0]!='#'] ) #remove '#' commented lines
            parser.parse_args(a.split(), namespace)
            namespace.input_filename = f.name
            setattr(namespace, self.dest, True)

def get_parser():
    parser = argparse.ArgumentParser(description='Craterstats: a tool to analyse and plot crater count data for planetary surface dating.')

    parser.add_argument('-i','--input', help="input args from file", type=open, action=LoadFromFile)
    parser.add_argument("-src", help="take command line parameters from text file", nargs='+', action=SpacedString)
    #latter is used for demo. Maybe switch demo to use .cs files, too?

    parser.add_argument("-lcs", help="list chronology systems", action='store_true')
    parser.add_argument("-lpc", help="list plot symbols and colours", action='store_true')
    parser.add_argument("-about", help="show program details", action='store_true')
    parser.add_argument("-demo", help="run sequence of demonstration commands: output in ./demo", action='store_true')

    parser.add_argument("-o","--out", help="output filename (omit extension for default) or directory", nargs='+', action=SpacedString)
    parser.add_argument("--functions_user", help="path to file containing user defined chronology systems", nargs='+', action=SpacedString)

    parser.add_argument("-f", "--format", help="output formats",  nargs='+', choices=['png','tif','pdf','svg','csv','stat'])

    parser.add_argument("-cs", "--chronology_system", help="chronology system index")
    parser.add_argument("-ef", "--equilibrium", help="equilibrium function index")
    parser.add_argument("-ep", "--epochs", help="epoch system index")

    parser.add_argument("-title", help="plot title", nargs='+', action=SpacedString)
    parser.add_argument("-pr", "--presentation", help="data presentation: " + (', ').join(cst.PRESENTATIONS))
    parser.add_argument("-xrange", help="x-axis range, log(min) log(max)", nargs=2)
    parser.add_argument("-yrange", help="y-axis range, log(min) log(max)", nargs=2)
    parser.add_argument("-isochrons", help="comma-separated isochron list in Ga, e.g. 1,3,3.7a,4a (optional combined suffix to modify label: n - suppress; a - above; s - small)")
    parser.add_argument("-show_isochrons", choices=['0', '1'], help="1 - show; 0 - suppress")
    parser.add_argument("-legend", help="0 - suppress; or any combination of: n - name, a - area, p - perimeter, c - number of craters, r - range, N - N(d_ref) value")
    parser.add_argument("-cite_functions", choices=['0','1'], help="1 - show; 0 - suppress")
    parser.add_argument("-mu", choices=['0','1'], help="1 - show; 0 - suppress")
    parser.add_argument("-style", choices=['natural', 'root-2'], help="diameter axis style")

    parser.add_argument("-invert", choices=['0','1'], help="1 - invert to black background; 0 - white background")
    parser.add_argument("-transparent", help="set transparent background", action='store_true')
    #combine invert/transparent into one? maybe not, but invert could be same syntax - get rid of 0,1
    parser.add_argument("-tight", help="tight layout", action='store_true')

    parser.add_argument("-pd", "--print_dimensions", help="print dimensions: either single value (cm/decade) or enclosing box in cm (AxB), e.g. 2 or 8x8")
    parser.add_argument("-pt_size", type=float, help="point size for figure text")
    parser.add_argument("-ref_diameter", type=float, help="reference diameter for displayed N(d_ref) values")
    parser.add_argument("-sf","--sig_figs", type=int, choices=[2,3], help="number of significant figures for displayed ages")

    parser.add_argument("-st","--sequence_table", help="generate sequence probability table", action='store_true')

    parser.add_argument("-p", "--plot", nargs='+', action=AppendPlotDict, metavar="KEY=VAL,",
                        help="specify overplot. Allowed keys:   \n"
                             "source=txt,"
                             "name=txt,"
                             "range=[min,max],"
                             "snap={1,0},"
                             "type={data,poisson,b-poisson,c-fit,d-fit},"
                             "error_bars={1,0},"
                             "hide={1,0},"
                             "colour={0-31},"
                             "psym={0-14},"
                             "binning={" +','.join(cst.Cratercount.BINNINGS) + "},"
                             "age_left={1,0},"
                             "display_age={1,0},"
                             "resurf={1,0}, apply resurfacing correction,"
                             "resurf_showall={1,0}, show all data with resurfacing correction,"
                             "isochron={1,0}, show whole fitted isochron,"
                             "offset_age=[x,y], in 1/20ths of decade")
    return parser

def defaults():
    set = {
        'chronology_system': 'Moon, Neukum (1983)',
        'cite_functions': 1,
        'epochs': '',
        'equilibrium': '',
        'invert': 0,
        'isochrons': '',
        'legend': 'nacr',
        'mu': 1,
        'presentation': 'differential',
        'print_dimensions': '7.5x7.5',
        'pt_size': 8.0,
        'randomness': 0,
        'ref_diam': 1,
        'sig_figs': 3,
        'show_isochrons': 1,
        'show_legend_area': 1,
        'show_title': 1,
        'style': 'natural',
        'title': '',
        'format': {'png', 'csv'}
        }
    plot = {
        'source': '',
        'name': '',
        'range': ['0', 'inf'],
        'snap': 1,
        'type': 'data',
        'error_bars': 1,
        'hide': 0,
        'colour': 0,
        'psym': 1,
        'binning': 'pseudo-log',
        'age_left': 0,
        'display_age': 1,
        'resurf': 0,
        'resurf_showall': 0,
        'isochron': 0,
        'offset_age': [0, 0]
    }
    return {'set':set,'plot':plot}


def decode_abbreviation(s,v,one_based=False,allow_ambiguous=False):
    """
    decode arbitrary abbreviation of list member into index

    :param s: full string list
    :param v: abbreviation or index
    :param one_based: menu indexing of s starts at 1?
    :param allow_ambiguous: if allowed, return first match
    :return: index (zero-based)
    """

    #perhaps add back index selection requiring # prefix?

    regex = '(?i)' + '.*'.join(v)
    res = [(i,e) for i,e in enumerate(s) if re.search(regex, e) is not None]
    res.sort(key=lambda x: len(x[1]))
    if len(res) == 0:
        sys.exit('Invalid abbreviation: ' + v)
    elif len(res) > 1 and not allow_ambiguous:
        sys.exit('Ambiguous abbreviation: ' + v)
    return res[0][0]


def construct_cps_dict(args,c,f,default_filename):
    if 'presentation' in vars(args):
        if args.presentation is not None:
            c['presentation'] = cst.PRESENTATIONS[decode_abbreviation(cst.PRESENTATIONS, args.presentation,one_based=True)]
    if c['presentation'] in ['chronology', 'rate', 'sequence']: #possible to overwrite with user-choice
        c['xrange'] = cst.DEFAULT_XRANGE[c['presentation']]
        c['yrange'] = cst.DEFAULT_YRANGE[c['presentation']]
    if c['presentation']=='sequence':
        c['legend']='A'
    #c['format'] = set(c['format']) if 'format' in c else {}

    for k,v in vars(args).items():
        if v is None:
            if k == 'out': c[k] = default_filename # don't set as default in parse_args: need to detect None in source_cmds
        else:
            if k in ('title',
                     'isochrons',
                     'show_isochrons',
                     'legend',
                     'print_dimensions',
                     'pt_size',
                     'ref_diameter',
                     'cite_functions',
                     'sig_figs',
                     'randomness',
                     'mu',
                     'invert',
                     'show_title',
                     'style',
                     'xrange', 'yrange',
                     ):
                c[k]=v
            elif k in ('chronology_system','equilibrium','epochs'):
                names = [e['name'] for e in f[k]]
                c[k]=f[k][decode_abbreviation(names, v, one_based=True, allow_ambiguous=True)]['name']

            elif k == 'out':
                if os.path.isdir(v):
                    c[k] = os.path.normpath(v+'/'+default_filename)
                else:
                    c[k] = gm.filename(v, 'pn')
                    ext = gm.filename(v, 'e').lstrip('.')
                    if ext: c['format'].add(ext)
            elif k == 'format':
                c[k]=set(v)


    cs=next((e for e in f['chronology_system'] if e['name'] == c['chronology_system']), None)
    if cs is None: sys.exit('Chronology system not found:' + c['chronology_system'])

    c['cf'] = cst.Chronologyfn(f, cs['cf'])
    c['pf'] = cst.Productionfn(f, cs['pf'])

    if 'equilibrium' in c and c['equilibrium'] not in (None,''):
        c['ef'] = cst.Productionfn(f, c['equilibrium'], equilibrium=True)
    if 'epochs' in c and c['epochs'] not in (None,''):
        c['ep'] = cst.Epochs(f, c['epochs'],c['pf'],c['cf'])

    if c['presentation'] == 'Hartmann':
        if hasattr(c['pf'],'xrange'): #not possible to overwrite with user choice
            c['xrange'] = c['pf'].xrange
            c['yrange'] = c['pf'].yrange
        else:
            c['xrange'] = cst.DEFAULT_XRANGE['Hartmann']
            c['yrange'] = cst.DEFAULT_YRANGE['Hartmann']

    return c


def construct_plot_dicts(args,plot):
    #if type(plot) is list: plot=plot[0] #take only first plot entry as template
    cpl = []
    specified_source = False
    if args.plot is None: return []
    for d in args.plot:
        p=plot.copy()
        if cpl: # for these items: if not given, carry over from previous
            for k in ['source','psym','snap','isochron','error_bars','colour','binning']:
                p[k] = cpl[-1][k]
            if p['source'] == cpl[-1]['source']: #only carry type if source unchanged
                p['type'] = cpl[-1]['type']

        for k0,v in d.items():
            k=cst.CRATERPLOT_KEYS[decode_abbreviation(cst.CRATERPLOT_KEYS, k0, allow_ambiguous=True)]
            if k in (
                    'name',
                    'snap',
                    'error_bars',
                    'hide',
                    'age_left',
                    'display_age',
                    'resurf',
                    'resurf_showall',
                    'isochron',
                    ):
                p[k]=v
            elif k in ('range','offset_age'):
                p[k] = v.strip('[]').split(',')
            elif k == 'source':
                v = v.replace('%sample%/', cst.PATH+'sample/')
                p['source']=v.strip('"')
                specified_source = True
            elif k == 'type':
                p[k]=cst.OPLOT_TYPES_SHORT[decode_abbreviation(cst.OPLOT_TYPES, v, allow_ambiguous=True)]
            elif k == 'binning':
                p[k] = cst.Cratercount.BINNINGS[decode_abbreviation(cst.Cratercount.BINNINGS, v, allow_ambiguous=True)]
                cst.Cratercount.BINNINGS
            elif k == 'colour':
                if v[0]=='#':
                    if bool(re.match(r'^#([0-9A-Fa-f]{6})$', v)):
                        p[k]=v
                    else:
                        sys.exit('Invalid colour:'+v)
                else:
                    names=[n for c1,c2,n in cst.PALETTE]
                    p[k]=decode_abbreviation(names, v, allow_ambiguous=True)
            elif k == 'psym':
                j=[i for i,e in enumerate(cst.MARKERS) if e[0]==v]
                if len(j)==1: # found standard abbreviation
                    p[k]=j[0]
                else: # look for arbitrary abbreviation or index
                    names = [e[1] for e in cst.MARKERS]
                    p[k]=decode_abbreviation(names, v, allow_ambiguous=True)

        if not specified_source: sys.exit('Source not specified')
        if os.path.isabs(p['source']):
            src = p['source']
        else:
            src = (gm.filename(args.input_filename,'p') if args.input else '')+p['source']
        p['cratercount'] = cst.Cratercount(src)
        cpl += [p]
    return cpl


def source_cmds(src):
    cmd=gm.read_textfile(src, ignore_blank=True, ignore_hash=True)
    parser=get_parser()
    for i,c in enumerate(cmd):
        a = c.split()
        args = parser.parse_args(a)
        if args.out is None:
            f='{:02d}-out'.format(i)
            a+=['-o',f]
            c+=' -o '+f
        print(f'\nCommand: {i}\ncraterstats ' + c)
        main(a)
    print('\nProcessing complete.')

def demo(d=None,src=cst.PATH+'config/demo_commands.txt'):
    cmd = gm.read_textfile(src, ignore_blank=True, ignore_hash=True)
    out='demo/'
    os.makedirs(out,exist_ok=True)
    f = '-o '+out+'{:02d}-demo '
    if d is None:
        d=range(0,len(cmd))
    for i in d:
        c=cmd[i]
        print(f'\nDemo {i}\ncraterstats '+c)
        a = (f.format(i) + c).split()
        main(a)
    print('\nDemo output written to: '+out)

def do_functions_user(args):
    if os.environ.get('CONDA_PREFIX'):
        config = os.environ.get('CONDA_PREFIX')+'/etc/config_functions_user.txt'
    else:
        return None #actions environment is not conda
    if args.functions_user:
        s=['#path to functions_user.txt',args.functions_user]
        gm.write_textfile(config,s)
    return gm.read_textfile(config,ignore_hash=True)[0] if gm.file_exists(config) else None

def main(args0=None):
    args = get_parser().parse_args(args0)
    if not args0: args0=sys.argv[1:]

    functions=cst.PATH+'config/functions.txt'
    s = gm.read_textfile(functions, ignore_hash=True, strip=';', as_string=True)  # read and remove comments
    functions_user = do_functions_user(args)
    if functions_user:
        try:
            s = s + gm.read_textfile(functions_user, ignore_hash=True, strip=';', as_string=True)
        except:
            print("Unable to read "+functions_user+" - ignoring.")
    functions = gm.read_textstructure(s,from_string=True)

    if args.lcs:
        print(gm.bright("\nChronology systems:"))
        print('\n'.join(['{0}'.format(e['name']) for e in functions['chronology_system']]))
        print(gm.bright("\nEquilibrium functions:"))
        print('\n'.join(['{0}'.format(e['name']) for e in functions['equilibrium']]))
        print(gm.bright("\nEpoch systems:"))
        print('\n'.join(['{0}'.format(e['name']) for e in functions['epochs']]))
        return

    if args.lpc:
        print(gm.bright("\nPlot symbols:"))
        print(', '.join(['{0} ({1})'.format(e[1], e[0]) for e in cst.MARKERS]))
        print(gm.bright("\nColours:"))
        print(', '.join(['{0}'.format(e[2]) for e in cst.PALETTE]))
        return

    if args.about:
        print('\n'.join(cst.ABOUT))
        return

    if args.src:
        source_cmds(args.src)
        return

    if args.demo:
        demo()
        return

    dflt = defaults()
    cp_dicts = construct_plot_dicts(args,dflt['plot'])
    if args.input:
        default_filename = gm.filename(args.input_filename,'pn')
    else:
        default_filename = '_'.join(sorted(set([gm.filename(d['source'], 'n') for d in cp_dicts]))) if cp_dicts else 'out'
    cps_dict = construct_cps_dict(args, dflt['set'], functions, default_filename)

    if 'a' in cps_dict['legend'] and 'b-poisson' in [d['type'] for d in cp_dicts]:
        cps_dict['legend']+='p' #force to show perimeter with area if using b-poisson


    cps=cst.Craterplotset(cps_dict) #,craterplot=cpl)
    for d in cp_dicts:
        if isinstance(d['colour'], int):d['colour']=cps.palette[d['colour']]
    cpl = [cst.Craterplot(d) for d in cp_dicts]
    cps.craterplot=cpl

    if cpl and cps.presentation not in ('sequence'):
        cps.autoscale(cps_dict['xrange'] if 'xrange' in cps_dict else None,
                      cps_dict['yrange'] if 'yrange' in cps_dict else None)

    if not args.input:
        gm.write_textfile(cps_dict['out']+'.cs',''.join(['\n'+e if e[0]=='-' and not (e+' ')[1].isdigit() else ' '+e for e in args0])[1:])

    drawn=False
    for functions in cps.format:
        if functions in {'png','pdf','svg','tif'}:
            if not drawn:
                cps.draw()
                drawn=True
            cps.fig.savefig(cps_dict['out']+'.'+functions, dpi=500, transparent=args.transparent,
                            bbox_inches='tight' if args.tight else None,pad_inches=.02 if args.tight else None)
        if functions in {'csv'}:
            cps.create_summary_table(f_out=cps_dict['out']+'.'+functions)
        if functions in {'stat'}:
            stat_files=set([(e.source,e.binning) for e in cpl])
            for stat in stat_files:
                cc=cst.Cratercount(stat[0])
                cc.WriteStatFile(gm.filename(stat[0],"n12", '_'+stat[1],".stat"),stat[1])


if __name__ == '__main__': 
    main()





#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import sys
import numpy as np
import scipy.optimize as sc

import craterstats as cst
import craterstats.gm as gm


class Productionfn:
    
    def __init__(self,source,identifier,equilibrium=False):
        
        self.name=identifier
        pf_type='equilibrium' if equilibrium else 'production'
        
        if type(source) is dict:
            src=source
        else:
            if '\n' in source: # multiline string is definition
                txt = source + '\n'+pf_type+'={\n name="null"\n}' # add null entry to force implied array
            else: # single line string is filename
                txt = gm.read_textfile(source, as_string=True, ignore_hash=True)
            src = gm.read_textstructure(txt, from_string=True)
        
        self.definition=next((e for e in src[pf_type] if e['name']==identifier), None)
        if self.definition is None:
            raise ValueError(pf_type.title()+' function not found: '+identifier)
        
        self.type=self.definition['type'] if 'type' in self.definition else 'polynomial'

        if 'range' in self.definition:
            self.range=np.float32(np.array(self.definition['range']))
        else:
            self.range = np.array([1e-4, 2e3])  # default max range

        if 'yrange' in self.definition: #for fixed Hartmann plots
            self.yrange=np.float32(np.array(self.definition['yrange']))

        if self.type=='polynomial':
            self.a=[float(e) for e in self.definition['coefficients']]
            self.C = self.polynomial_C
            self.C10 = self.polynomial_C10
            self.F = self.polynomial_F
            self.F10 = self.polynomial_F10
            self.R = self.polynomial_R
            self.H = self.polynomial_H
            
        elif self.type=='tabular incremental root-2':
            h=[float(e) for e in self.definition['H']]
            index_1km=int(self.definition['index_1km'])
            hC10=np.log10(list(reversed(np.cumsum(list(reversed(h))))))
            self.hC10=hC10-hC10[index_1km]

            d_min=np.array([2**((i-index_1km)/2) for i,e in enumerate(h)])
            d_mean=d_min*2**.25
            self.d_min10=np.log10(d_min)
            self.d_mean10=np.log10(d_mean)
            self.range=d_mean[[0,-1]]
            self.xrange = np.log10([d_min[0],d_min[-1]*2**.5]) # plot axis range

            self.C = self.hartmann_C
            self.C10 = self.hartmann_C10
            self.F = self.hartmann_F
            self.F10 = self.hartmann_F10
            self.R = self.hartmann_R
            self.H = self.hartmann_H


    def __str__(self):
        return self.name


    def evaluate(self,presentation,d,a0=0.):
        '''
        Evaluate PF according to presentation string

        :param presentation: presentation string
        :param d: series of crater diameters (km)
        :param a0: a0
        :return: y-values
        '''
        if presentation=='cumulative':
            return self.C(d,a0)
        elif presentation=='differential':
            return self.F(d,a0)
        elif presentation=='R-plot':
            return self.R(d,a0)
        elif presentation=='Hartmann':
            return self.H(d,a0)
        
           
#polynomial pf

    def polynomial_C10(self, d10, a0):
        a = [a0] + self.a[1:]
        C10 = gm.poly(a, d10)
        return C10

    def polynomial_C(self,d,a0): #cumulative fn
        x=np.log10(d)
        C=10**self.polynomial_C10(x,a0)
        return C

    def polynomial_F10(self, d10, a0):
        F10 = np.log10(self.polynomial_F(10**d10,a0))
        return F10
    
    def polynomial_F(self,d,a0): #differential form
        x=np.log10(d)
        a=[a0]+self.a[1:]
        p= gm.poly(a, x)
        dCdp=np.log(10.)*10.**p
        a_dash=[(i+1)*e for i,e in enumerate(a[1:])]
        dpdx= gm.poly(a_dash, x)
        dxdD=1./(d*np.log(10.))    
        F=-dCdp*dpdx*dxdD  #(Michael & Neukum 2010, Eq 1)
        return F
    
    def polynomial_R(self,d,a0): #relative (R) (Arvidson et al, 1979)
        F=self.polynomial_F(d,a0)  
        R=F/(d**(-3.))          
        return R
    
    def polynomial_H(self,d,a0): #Hartmann diagram 
        F=self.polynomial_F(d,a0)
        H=F*d*(2**.25-2**(-.25)) #adjust to bin minimum (Michael 2013, Eq. 1)
        return H

#Hartmann tabular incremental root-2

    def hartmann_C10(self,d10,a0): #cumulative fn
        return self.hartmann_calc(d10,a0,'C10',log10=True)

    def hartmann_C(self,d,a0): #cumulative fn
        return self.hartmann_calc(d,a0,'C')
        
    def hartmann_H(self,d,a0): #Hartmann diagram 
        return self.hartmann_calc(d,a0,'H')
    
    def hartmann_F10(self,d10,a0): #differential form
        return self.hartmann_calc(d10,a0,'F10',log10=True)

    def hartmann_F(self,d,a0): #differential form
        return self.hartmann_calc(d,a0,'F')

    def hartmann_R(self,d,a0): #relative (R) (Arvidson et al, 1979)  
        return self.hartmann_calc(d,a0,'R')
    
    def hartmann_calc(self,x,a0,mode,log10=False):
        #C,H,F,R - lower case represent continuous functions; upper case - evaluations for given d
        d10=x if log10 else np.log10(x)
        c=10**(self.hC10+a0)
        
        if mode in ('C','C10'):
            C10 = np.interp(d10, self.d_min10, np.log10(c))
            return 10**C10 if mode=='C' else C10

        h=c-np.concatenate((c[1:],[0]))
        
        if mode=='H':
            H=10**np.interp(d10,self.d_mean10,np.log10(h))
            return H
        
        beta=np.sqrt(2)
        bin_width0=2**.25-2**-.25
        f_bin=h/(10**self.d_mean10)/bin_width0

        k = (self.hC10 - np.roll(self.hC10,-1)) / np.log10(beta)
        k[-1]=k[-2] #fix slope at end to be continuous
        bc=cst.bin_bias_correction(beta, k) #remove binning bias
        f=f_bin/bc       
        
        if mode in ('F','F10'):
            F10=np.interp(d10,self.d_mean10,np.log10(f))
            return 10**F10 if mode=='F' else F10
        
        r=f/(10**self.d_mean10)**-3
        
        if mode=='R': 
            R=10**np.interp(d10,self.d_mean10,np.log10(r))
            return R


    def fit(self,p):
        '''
        Make fit to points supplied in dict, according to presentation (cumulative or differential)

        :param p: dict containing presentation, y, d and err fields
        :return: a0 tuple: (fit, lower, upper)
        '''
        y, d, err = p['y'], p['d'], p['err']

        if type(y) is np.ndarray:
            if y.size == 0:
                sys.exit('Cannot fit to empty range')
            if y.size==1:
                y,d,err=y[0],d[0],err[0] # must convert to float so that next line works for float or ndarray
        if isinstance(y, (np.floating, float)): #allow to call with single values for conversion
            y,d,err=np.array([y,y]),np.array([d,d]),np.array([err,err])

        with np.errstate(divide='ignore'):
            if p['presentation'] == 'cumulative':
                func, sigma = self.C10, 1 / np.sqrt(y)
            elif p['presentation'] == 'differential':
                func, sigma = self.F10, 1 / np.sqrt(err)

        popt, pcov = sc.curve_fit(func, np.log10(d), np.log10(y),sigma=sigma)
        a0=popt[0]-[0,np.log10(y[0])-np.log10(y[0]-.98*err[0]),np.log10(y[0])-np.log10(y[0]+.98*err[0])]
        return a0


    def getplotdata(self,presentation,a0=None,range=None):
        '''
        Return dict of values for plotting PF

        :param presentation: presentation string
        :param a0: a0
        :param range: diameter range
        :return: dict with series of values for plotting
        '''
        ns=400
        if a0 is None: a0 = self.a[0]
        if range is None: range=self.range
        range=np.clip(range,self.range[0],self.range[1])
        log_d=np.linspace(np.log10(range[0]),np.log10(range[1]),ns)
        d=10.**log_d
        y=self.evaluate(presentation,d,a0)
        return {'presentation':presentation,'d':d,'y':y}


    def getisochron(self,presentation,a0,ef):
        '''
        Return series of values for plotting isochron, optionally truncated by equilibrium function curve

        :param presentation: presentation string
        :param a0: a0
        :param ef: equilibrium function instance (optional)
        :return: dict with series of values for plotting
        '''

        iso=self.getplotdata(presentation,a0)
        if ef:
            e=ef.getplotdata(presentation)
            j= gm.value_locate(e['d'], iso['d'])
            q= gm.where(iso['y'] < e['y'][j] * .8)
            iso={'presentation':presentation,'d':iso['d'][q],'y':iso['y'][q]}

        return iso



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

def poly(p,x):
    """Evaluate polynomial with coefficients in ASCENDING order."""

    p1=p[::-1]
    return np.polyval(p1,x)


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

from . import normal


def poisson(k,lam,cumulative=False,threshold=23):
    '''
    poisson mass function
    vectorised in lambda, but not k

    :param k:
    :param lam:
    :param cumulative: return cmf instead of pmf
    :param threshold: use normal approximation above threshold value of k
    :return: pmf
    '''

    if k<threshold:       # poisson
        if cumulative:
            res = 0.
            for i in range(k+1):
                res += (lam**i) / np.math.factorial(i)
            return np.exp(-lam) * res
        else:
            return lam**k * np.exp(-lam) / np.math.factorial(k)

    else:                 # normal
        if cumulative:
            return normal(lam, np.sqrt(lam), k, cumulative=True)
        else:
            return normal(lam, np.sqrt(lam), k)




#
# import matplotlib.pyplot as plt
#
# if __name__ == '__main__':
#
#     ns=1000
#     x=np.linspace(0,100,ns)
#     fig, (ax1, ax2) = plt.subplot(1, 2)
#     for i in range(1,99):
#         ax1.plot(x,poisson(i,x),color='0' if i<30 else '0.')
#     plt.show()






#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import re

from ..string.strip_quotes import strip_quotes
from ..string.quoted_split import quoted_split
from . import read_textfile


def simple_value(s):
    '''evaluate single (possibly multi-line) value'''
  
    if s.startswith('*'): #no need for pointer, since using lists: just ignore
        s=s[1:]
    
    if s.startswith('['):
        w=s.strip('[] \n').split(',')
        r=[strip_quotes(e.strip('\n ')) for e in w]
        
    elif s.startswith('('):   
        w=s.strip('() \n').split(',')
        r=[strip_quotes(e.strip('\n ')) for e in w]
    
    elif s.startswith('{'):  #declare ascii table: should be field names after (otherwise, was structure)
        w=s.splitlines()
        tag0=w[0].split(',')      
        tag=[e.strip('{ ') for e in tag0]      
        
        r={t:[] for t in tag}
        for row in w[1:]:
            v=quoted_split(row)
            if v[0]=='}': break
            for t in tag:
                r[t].append(v.pop(0) if len(v)>0 else '')
    else: 
        r=strip_quotes(s.strip(' \n'))
    
    return r


def merge(s1, sr):
    '''merge single tag structure into cumulative structure'''

    # if first is duplicate of tag in remainder (implicit array) -> add to array
    # else make union of first+remainder
    
    if type(sr) != dict: return s1
    s1_tag=list(s1)[0]
    if s1_tag in list(sr):
        if type(sr[s1_tag]) != list:
            sr.update({s1_tag:[s1[s1_tag]]+[sr[s1_tag]]})    
        else:
            sr[s1_tag].insert(0,s1[s1_tag]) 
    else:
       sr.update(s1) 
       
    return sr    


def evaluate(s):
    '''evaluate first definition; join with recursively evaluated remainder'''

    # find keyword followed by either struct_value, multiline text_value, or simple value; then following keyword
    m=re.search("(?ms)^\s*(?P<keyword>[\w\.]*\^?)\s*=\s*"
                "(\{\s*$(?P<struct_value>[^{]*)^\s*\}.*?|"
                "\"\s*$(?P<text_value>[^{]*)^\s*\".*?|"
                "(?P<value>.*?))"
                "^\s*(?P<keyword2>[\w\.]*\^?) *="
                ,s+"\nend=")
      
    if not m: return None #no definitions to process
    
    keyword=m['keyword']
    if keyword.endswith('^'): keyword='ptr_'+keyword.rstrip('^')
    
    #if structure, send contents for evaluation
    if m['struct_value'] != None:  
        r=evaluate(m['struct_value'])
       
    elif m['text_value'] != None: #text block
        r=m['text_value']
    
    #if simple value, evaluate directly      
    else: r=simple_value(m['value']) #simple value
    
    #new entry
    dict1={keyword:r}     
    
    #evaluate remainder  
    remainder = s[m.start('keyword2'):] if m['keyword2'] != None else ''    
    dict_r=evaluate(remainder)
    
    #join together  
    dict_out=merge(dict1, dict_r)
    
    return dict_out #structure from s

def read_textstructure(p,from_string=False):
    '''
    Read set of key-value pairs from text file

    :param p: filename
    :param from_string: interpret string in p directly
    :return: dictionary
    '''
    
    if not from_string:
        s=read_textfile(p, ignore_hash=True, strip=';', as_string=True) #read and remove comments
    else:
        s=p
        
    return evaluate(s)






















































#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np
import types
import craterstats.gm as gm


def import_code(code):
    module = types.ModuleType('cf')
    exec(code, module.__dict__)
    return module


class Chronologyfn:
    
    def __init__(self,source,identifier):
        
        t_steps=1000
        self.lambda_n1=None
        self.lambda_phi=None

        self.name=identifier
        
        if type(source) is dict:
            src=source
        else:
            if '\n' in source: # multiline string is definition
                txt = source + '\nchronology={\n name="null"\n}' # add null entry to force implied array
            else: # single line string is filename
                txt = gm.read_textfile(source, as_string=True, ignore_hash=True)
            src = gm.read_textstructure(txt, from_string=True)
        
        self.definition=next((e for e in src['chronology'] if e['name']==identifier),None)
        if self.definition is None:
            raise ValueError('Chronology function not found: '+identifier)
              
        if 'coefficients' in self.definition:            
            p=[float(e) for e in self.definition['coefficients']]
            self.lambda_n1 = lambda t: p[0]*(np.exp(p[1]*t)-1.)+p[3]*t
            self.lambda_phi = lambda t: p[1]*p[0]*np.exp(p[1]*t)+p[3]  
      
        if 'n1_code' in self.definition:
            # '*' allows unforseen fns to be used in user code without np prefix (or user-awareness of python)
            code='from numpy import *\n' \
                'def n1(t):\n' \
                +'\n'.join(['  '+e.strip() for e in self.definition['n1_code'].splitlines()]) \
                +'\n  return n1'
            m=import_code(code)
            self.lambda_n1=m.n1
            
        self.ts=np.linspace(0.,5.,t_steps)
        self.n1s=self.N1(self.ts)     
        
    def __str__(self):
        return self.name
    
    def N1(self,t):  #t in Ga
        return self.lambda_n1(t)

    def a0(self,t):
        return np.log10(self.N1(t))
        
    def phi(self,t): #phi = dN1/dt - the impact rate [/Ga]
        if self.lambda_phi is None:
            dt=t/1e4
            return (self.N1(t+dt)-self.N1(t-dt))/(2.*dt)
        else:
            return self.lambda_phi(t)
        
    def t(self,a0=None,n1=None):
        if not a0 is None:
            n1=10**a0
        
        t=np.interp(n1,self.n1s,self.ts)
        return t
             
    def getplotdata(self,phi=False,linear=False):
        ns=1000
        t=np.linspace(0.,5.,ns) if linear else 10**np.linspace(-9,np.log10(5),ns) #maybe could reuse self.t (switch to log there?)
        y=self.phi(t) if phi else self.N1(t)
        return t,y









#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import re

def read_textfile(filename,n_lines=-1,ignore_blank=False,ignore_hash=False,strip=None,as_string=False):
    '''

    :param filename: full filepath
    :param n_lines: number of lines to read (all, if omitted)
    :param ignore_blank: ignore blank lines
    :param ignore_hash: ignore # lines
    :param strip: remove trailing comments after strip symbol (e.g. ';')
    :param as_string: concatenate into single string, joined with '\n'
    :return: file contents as list of strings or single string
    '''

    with open(filename, 'r', encoding='utf-8-sig') as file: # encoding='utf-8-sig' removes BOM if present
        if n_lines !=-1:
            s=[]
            for i in range(n_lines): s.append(file.readline().strip())
        else:
            s=file.read().splitlines()

    if ignore_blank:
        s = [e for e in s if e != '']

    if ignore_hash:
        s = [e for e in s if not e.lstrip().startswith('#')]

    if strip is not None:
        c = re.compile(r'(("[^"]*?")|(\'[^\']*?\')|[^;\'"]*)*') # avoid splitting within quoted string
        s = [c.match(e)[0] for e in s]

    if as_string:
        s='\n'.join(s)

    return s






#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

def where(e):
    if type(e) is np.ndarray:
        res=np.where(e)
        return list(res[0])




#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import re

def filename(f,code,insert1=None,insert2=None,max_ext_length=4):
    '''
    Split path, filename, extension into constituent parts and reconstruct as required

    :param f: file name or full path
    :param code: string containing any order/combination of
        p - path; n - name (without extension); e (extension with '.'); 1,2 - insert string;
        u - up (parent path); b - back (parent dir name)
    :param insert1,insert2: optional strings to insert (if arrays, should agree in dim with f)
    :param max_ext_length: max extension length (so that files including dots but no extension are not cut)
    :return: reconstructed filename or fragment(s)

    e.g. filename(r"d:\tmp\fred.txt","pn1e","_en") gives "d:\tmp\fred_en.txt"
    e.g. filename(r"d:\tmp\fred.txt","ne") gives "fred.txt"
    e.g. filename(r"d:\tmp\fred.txt","bne") gives "tmp/fred.txt"
    e.g. filename(r"F:\broom\ortho-090\nd.l3_he330_0000.tif","p1bne",'nd\\') gives 'F:\\broom\\nd\\ortho-090\\nd.l3_he330_0000.tif'
    '''

    if type(f) is list:
        return [filename(e,code,insert1=insert1,insert2=insert2) for e in f]

    m = re.search(r'(?P<path>.*?)(?P<name>[\.,= \w-]*?)(?P<ext>(\.\w{1,'+str(max_ext_length)+'})?$)', f)
    path=m['path']
    name=m['name']
    ext = m['ext']

    out=''
    for ch in code[::-1]:
        if ch=='p': out=path+out
        elif ch=='n': out=name+out
        elif ch=='e': out=ext+out
        elif ch=='1': out=insert1+out
        elif ch=='2': out=insert2+out
        elif ch in ['u','b']:
            m = re.search(r'(?P<path>.*(\\|/))(?P<parentdir>.+)', path)
            path=m['path'] if m['path'] is not None else ''
            parentdir=m['parentdir'] if m['parentdir'] is not None else ''
            out=(parentdir if ch=='b' else path)+out

    return out






#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.
import copy

import numpy as np
import sys

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.figure as mfig

import craterstats as cst
import craterstats.gm as gm


class Craterplotset:
    """
    Base object for set of Craterplots. Holds properties common to all of set.
    Creates plot layout and calls overplot methods of Craterplot instances.
    Aggregates legend annotations.

    """
    
    def __init__(self,*args,**kwargs):   
            
        # matplotlib objects
        self.fig=None
        self.ax=None
        
        self.craterplot=[]  #craterplot layers   
        self.cf=None        #chronology fn
        self.pf=None        #production fn
        self.ef=None        #equilibrium fn
        self.ep=None        #epoch system

        self.cm2inch=1/2.54

        self.UpdateSettings({
            'title':'',
            'presentation':'differential',
            'xrange':[-3,2],
            'yrange':[-5,5],
            'style':'natural',
            'isochrons':'',
            'show_isochrons':0,
            'legend_data':'a', # a show counting area
            'legend_fit':'r', # n - show N(d_ref); r - show diam range
            'print_dimensions':'7.5x7.5',
            'pt_size':9.,
            'ref_diameter':1.,
            'cite_functions':1,
            'sig_figs':3,
            'randomness':0,
            'mu':1,
            'invert':0,
            'show_title':1,
            },*args,kwargs)

        self.sz_ratio= self.pt_size / 9.
        self.marker_def = [e[2].copy() for e in cst.MARKERS]
        for e in self.marker_def:
            e['markersize'] *= self.sz_ratio
            e['markeredgewidth'] = .5 * self.sz_ratio
        self.grey = cst.GREYS[self.invert]
        self.palette = [e[self.invert] for e in cst.PALETTE]

        
    def UpdateSettings(self,*args,**kwargs): #pass either dictionary or keywords
        a = {k: v for d in args for k, v in d.items()}
        a.update(**kwargs)
        for k, v in a.items():
            #if k == 'source': self.cratercount = cst.Cratercount(v)
            if k in ('xrange','yrange'): v=[float(e) for e in v]
            if k in ('pt_size','ref_diameter'): v = float(v)
            if k in ('show_isochrons',
                    'cite_functions',
                    'randomness',
                    'mu',
                    'invert',
                    'show_title',
                    'sig_figs',
                     ):
                v = int(v)
            setattr(self, k, v)


    def EstablishFontandScaling(self):
        # set up font and scaled font size

        plt.style.use(('default', 'dark_background')[self.invert])

        desired_font = ['Myriad Pro', 'Verdana', 'DejaVu Sans', 'Tahoma']  # in order of preference
        available_font = gm.mpl_check_font(desired_font)
        scale_factor = {'Myriad Pro': 1.,
                       'Verdana': .83,
                       'DejaVu Sans': .83,
                       'Tahoma': .93,
                       }[available_font]
        self.scaled_pt_size = self.pt_size * scale_factor

        plt.rc('font', family=available_font, size=self.scaled_pt_size)
        plt.rc('mathtext', fontset='custom', rm=available_font, bf=available_font + ':bold',
              cal=available_font + ':italic')

        tw = .4 * self.sz_ratio
        plt.rcParams.update({
            'legend.fontsize': 'x-small',
            'legend.title_fontsize': 'small',
            'axes.labelsize': self.scaled_pt_size * .9,
            'axes.titlesize': 'small',
            'xtick.labelsize': 'medium',
            'ytick.labelsize': 'medium',
            'axes.titlepad': self.scaled_pt_size * 1.,
            'axes.linewidth': tw,
            'xtick.major.width': tw,
            'xtick.minor.width': tw,
            'ytick.major.width': tw,
            'ytick.minor.width': tw,
        })

            
    def CreatePlotSpace(self):
        """
        Set up plot dimensions, font and scaled font size, titles, tick marks, and tick labels
        Create layout
        Set up plot coordinate transformations

        """

        if self.fig: del self.fig
        self.EstablishFontandScaling()
            
#set up plot dimensions

        def f(x): return np.clip(gm.mag(x), 1, None)
        self.decades=f(self.xrange),f(self.yrange)

        self.data_aspect=(
            2                                 if self.presentation=='differential' else
            self.decades[1]/self.decades[0]   if self.presentation in ['Hartmann','chronology', 'rate'] else
            1
            )            
            
        plot_dims=[float(e) for e in self.print_dimensions.split('x')]
        
        self.print_scale=(                                                      #in cm/decade
            np.clip(plot_dims[0],1,None) if len(plot_dims)==1 else
            plot_dims[0]/self.decades[0] if self.presentation=='Hartmann' else
            min(np.clip(plot_dims*np.array([1,self.data_aspect]),1,None)/self.decades)
            )
                
        randomness_plot=int(self.randomness and not self.presentation in ['chronology', 'rate']) #boolean as 0,1
        margin=self.pt_size*.2  #plot margin in cm
        xsize=self.decades[0]*self.print_scale+2*margin #in cm

        ysize=(
            xsize+self.print_scale*randomness_plot  if self.presentation=='Hartmann' else
            xsize                                   if self.presentation in ['chronology', 'rate'] else
            (randomness_plot+self.decades[1]/self.data_aspect)*self.print_scale+2*margin
            )
              
        self.position=margin*np.array([1,1,xsize/margin-2,ysize/margin-2-randomness_plot]) #cm
        position_randomness=margin*np.array([1,ysize/margin-.8-randomness_plot,xsize/margin-2,ysize/margin-2]) #cm - not yet used
        normalised_position=self.position/np.array([xsize,ysize,xsize,ysize])  #pos, width for matplotlib axes



# set up titles, tick marks, and tick labels

        nonroot2=False in [cp.binning=='root-2' for cp in self.craterplot]

        xtitle="Age, Ga" if self.presentation in ['chronology', 'rate'] else 'Diameter'
        ytitle={
            'cumulative':   'Cumulative crater density, km$^{-2}$',           
            'differential': 'Differential crater density, km$^{-3}$',
            'R-plot': 'Relative crater density (R)',
            'Hartmann':     'Craters per √2-bin'+(' (equivalent)' if nonroot2 else '')+', km$^{-2}$',
            'chronology':   'Cumulative crater density N(>'+format(self.ref_diameter,'.0f')+' km), km$^{-2}$',
            'rate':         'Crater formation rate N(>'+format(self.ref_diameter,'.0f')+' km), km$^{-2}$ Ga$^{-1}$',
            }[self.presentation]
        
        title = ('\n'.join(self.title.split('|')) if self.show_title else None)

        xminor=xmajor=xtickv=None
        add_xminorlogticks=False

 
        if self.presentation in ['chronology', 'rate']:
            xminor, xmajor = .5, 1
        else:
            if self.style=='root-2' or self.presentation=='Hartmann':                
                v=[np.log10(2.**e) for e in range(-10,11)] #set up root-2/hartmann axis labels
                labels=[str(e) for e in
                        (1, 2, 4, 8, 16, 31, 63, 125, 250, 500, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024)]
                labels[10]='  '+labels[10]+"km"
                xtickv,xtickname = map(list,zip(*[(val,txt) for val,txt in zip(v,labels) if val>=self.xrange[0] and val<=self.xrange[1]]))
                if xtickv[0]<0: xtickname[0]+='m'
                xminor=(xtickv[1]-xtickv[0])/2
                plt.rcParams.update({'xtick.labelsize':self.scaled_pt_size*.75})
                
            elif self.style=='natural':                
                v=np.arange(np.ceil(self.xrange[0]),np.floor(self.xrange[1])+1).tolist()
                xtickname=[format(10**e,'0g')+'$\,$km' if e>=0 else format(10**(e+3),'0g')+'$\,$m' for e in v]
                xtickv=v
                add_xminorlogticks=True

#create layout

        fig = mfig.Figure(figsize=[xsize*self.cm2inch,ysize*self.cm2inch],dpi=200) # bypass pyplot figure manager (causes issues, and not needed for non-interactive)

        ax=fig.add_axes(normalised_position)
        ax.set_axisbelow(False) # draw axis on top of other elements
        
        if add_xminorlogticks:                      #draw minor xlog ticks using different axes, since main x axis is linear
            ax2=fig.add_axes(normalised_position,frameon=False)
            ax2.set_axisbelow(False)
            ax2.set_xscale('log')      
            ax2.set_xbound(lower=10**self.xrange[0], upper=10**self.xrange[1])  
            ax2.tick_params(axis='x',which='minor',direction="in",length=self.pt_size*.25, top=True)
            ax2.tick_params(axis='x',which='major', bottom=False,labelbottom=False)
            ax2.get_yaxis().set_visible(False)            
                    
        ax.set_yscale('log')                                             #don't use xlog scale - awful to set up ticks for Hartmann binning
        ax.set_ylim(bottom=10**self.yrange[0], top=10**self.yrange[1])   #nb craterstats2 used linear for both axes
  
        ax.set_xlim(left=self.xrange[0], right=self.xrange[1])
        
        ax.tick_params(axis='both',which='major',direction="in",length=self.pt_size*.5,right=True,top=True)
        ax.tick_params(axis='both',which='minor',direction="in",length=self.pt_size*.25,right=True,top=True)
 
        if xminor is not None: ax.xaxis.set_minor_locator(ticker.MultipleLocator(xminor))            
        if xmajor is not None: ax.xaxis.set_major_locator(ticker.MultipleLocator(xmajor))

        ax.yaxis.set_major_locator(ticker.LogLocator(base=10, numticks=self.decades[1]+2))
        ax.yaxis.set_minor_locator(ticker.LogLocator(base=10., subs=(.1,.2,.3,.4,.5,.6,.7,.8,.9), numticks=self.decades[1]+2))
        ax.yaxis.set_minor_formatter(ticker.NullFormatter())
                    
        if xtickv is not None:
            ax.set_xticks(xtickv)
            ax.set_xticklabels(xtickname)                
        
        if title:ax.set_title(title)
        if ytitle != '':ax.set_xlabel(xtitle)
        if ytitle != '':ax.set_ylabel(ytitle)

        self.fig=fig
        self.ax=ax

# set up coordinate transformations
        a2d=ax.transAxes + ax.transData.inverted()
        self.axis_to_data = a2d.transform
        self.data_to_axis = a2d.inverted().transform
        self.axis_to_fig = (ax.transAxes + fig.transFigure.inverted()).transform
     

    def create_sequence_plotspace(self):
        """
        Set up and draw sequence plot
        """

        if self.fig: del self.fig
        self.EstablishFontandScaling()

        # set up plot dimensions

        margin = self.pt_size * .2  # plot margin in cm
        xsize, ysize = [float(e)+2*margin for e in self.print_dimensions.split('x')]


        self.position = margin * np.array([1, 1, xsize / margin - 2, ysize / margin - 2])  # cm

        def f(x): return np.clip(gm.mag(x), 1, None)
        self.decades=f(self.xrange),f(self.yrange)

        self.t_max = np.clip(max(self.xrange), 1e-8, 4.5)
        self.t_min = np.clip(min(self.xrange), 0., 3.5)

        if self.t_min<1 and self.t_min !=0:
            self.t_min=10**np.floor(np.log10(self.t_min))
            self.crossover = True
        else:
            self.crossover = False


        if self.crossover:
            t_crossover = 3.
            log_t_min = np.log10(self.t_min)

            dec2lin = 0.7
            lin_units = max(self.t_max - t_crossover, 0) / dec2lin
            log_units = np.log10(t_crossover) - log_t_min

            xtickv = [3., 3.5, 4.]

        else:
            t_crossover = self.t_min
            lin_units = 1
            log_units = 0
            xtickv = [e/2. for e in range(0,10) if self.t_min <= e/2. <= self.t_max]

        xticklabels = [str(e) for e in xtickv]
        xticklabels[-1] += ' Ga'


        xfrac_linear = lin_units / (lin_units + log_units)
        width_linear = xfrac_linear * (xsize - margin * 2)
        pos_linear = [margin, margin, width_linear, ysize - 2*margin]
        pos_log = [margin + width_linear, margin, (1 - xfrac_linear) * (xsize - margin * 2), pos_linear[3]]
        pos_all = [margin, margin, xsize - margin * 2, ysize - margin * 2]

        fig = mfig.Figure(figsize=[xsize * self.cm2inch, ysize * self.cm2inch], dpi=500)

        ax_lin = fig.add_axes(pos_linear / np.array([xsize, ysize, xsize, ysize]))
        ax_log = fig.add_axes(pos_log / np.array([xsize, ysize, xsize, ysize]))
        ax = fig.add_axes(pos_all / np.array([xsize, ysize, xsize, ysize]))

        #ax_log
        if self.crossover:
            ax_log.set_xscale('log')
            ax_log.set_xlim(left=t_crossover, right=10 ** log_t_min)
            def xlogFuncFormatter(x, pos):
                return cst.str_age(x, simple=True)

            ax_log.xaxis.set_major_formatter(ticker.FuncFormatter(xlogFuncFormatter))

            ax_log.xaxis.set_major_locator(ticker.LogLocator(numticks=999))
            ax_log.xaxis.set_minor_locator(ticker.LogLocator(numticks=999, subs="auto"))

            ax_log.tick_params(length=2, pad=1.5)
            ax_log.tick_params(which='minor',length=1)

            ax_log.get_yaxis().set_visible(False)

            ax_log.spines[['left', 'right', 'top']].set_visible(False)
            ax_log.set_ylim(bottom=0, top=1)

            ax_log.plot(t_crossover,0,marker='x',clip_on=False,markeredgewidth=.3, markersize=self.pt_size*.3,color='black')
        else:
            ax_log.set_axis_off()


        #ax_lin
        ax_lin.set_xlim(left=self.t_max, right=t_crossover)
        ax_lin.set_ylim(bottom=0, top=1)


        ax_lin.set_xticks(xtickv, labels=xticklabels)
        ax_lin.tick_params(length=2, pad=1.5)
        ax_lin.tick_params(which='minor',length=1)
        ax_lin.xaxis.set_minor_locator(ticker.AutoMinorLocator(5))

        ax_lin.get_yaxis().set_visible(False)
        ax_lin.spines[['left', 'right', 'top']].set_visible(False)

        #ax_all
        ax.set_xlim(left=0, right=1)
        ax.set_ylim(bottom=0, top=1)
        ax.get_yaxis().set_visible(False)
        ax.get_xaxis().set_visible(False)
        ax.spines[['left', 'right', 'top', 'bottom']].set_visible(False)
        ax.patch.set_facecolor('none')

        title = ('\n'.join(self.title.split('|')) if self.show_title else None)
        if title: ax.set_title(title)

        # create layout

        self.fig = fig
        self.ax = ax
        self.ax2 = (ax_lin,ax_log)

        # set up coordinate transformations
        a2d = ax.transAxes + ax.transData.inverted()
        self.axis_to_data = a2d.transform
        self.data_to_axis = a2d.inverted().transform
        self.axis_to_fig = (ax.transAxes + fig.transFigure.inverted()).transform


    def draw(self):
        """
        Draw plot, adding all specified items

        """

        if self.presentation=='sequence':
            self.create_sequence_plotspace()
            if self.ep: self.ep.sequence_oplot(self)
        else:
            self.CreatePlotSpace()

        N=self.pf.evaluate("cumulative",[self.ref_diameter,1.]) if self.pf else [1,1]
        ref_diam_ratio=N[0]/N[1]

        if self.cite_functions:
            txt=''
            if self.ep: txt += "Epochs: " + self.ep.name + '\n'
            if self.ef and self.presentation not in ['sequence']: txt += "EF: " + self.ef.name + '\n'
            if not (self.presentation in ['chronology', 'rate'] and self.ref_diameter==1) or self.ep: txt += "PF: " + self.pf.name + '\n'
            txt += "CF: " + self.cf.name
            #offset would be better specified in pts (1.5 * fontsize) [for many-decade plots]:
            self.ax.text(.05, .05, txt, transform=self.ax.transAxes, fontsize=self.scaled_pt_size * .7, linespacing=1.5)

        if self.presentation in ['chronology', 'rate']:
            phi=self.presentation=='rate'
            if self.ep:
                self.ep.chronology_oplot(self,phi=phi)
            t,y=self.cf.getplotdata(linear=True,phi=phi)

            if self.ref_diameter != 1.:  # only show legend if not N(1)
                self.ax.plot(t, y * ref_diam_ratio, label='N(>{:g} km)'.format(self.ref_diameter),
                             lw=1, color=self.palette[0], marker=None)
                self.ax.plot(t, y, label='N(>1 km)', color=self.grey[0], lw=1, marker=None)
            else:
                self.ax.plot(t, y, color=self.palette[0], lw=1, marker=None)

        if self.presentation in ['cumulative', 'differential', 'R-plot', 'Hartmann']:
            if self.ep: #overplot epochs
                self.ep.oplot(self)
            if self.ef:
                e = self.ef.getplotdata(self.presentation)
                self.ax.plot(np.log10(e['d']), e['y'], color=self.grey[0], lw=.7)

            if self.show_isochrons and self.isochrons:
                self.plot_isochrons()

        if self.presentation in ['cumulative', 'differential', 'R-plot', 'Hartmann', 'sequence']:
            for cp in self.craterplot:
                cp.overplot(self)

        # aggregate legend entries
        h, b = self.ax.get_legend_handles_labels()
        h1, b1 =[], []
        skip=False
        for i,e in enumerate(b):
            if skip:
                skip=False
                continue
            if e=='fit':
                skip = True
                if set(self.legend) & set('crN') != set():
                    h1.append((h[i],h[i+1]))
                    b1.append(b[i+1])
            else:
                h1.append(h[i])
                b1.append(e)

        self.ax.legend(h1,b1,frameon=False, loc='upper right', borderaxespad=1,handletextpad=0.5,labelspacing=.4,handlelength=1.5)


    def plot_isochrons(self):
        """
        Overplot predefined isochrons with annotations

        :return: none
        """

        isochrons = [(abs(float(e.rstrip('ash'))),'h' in e,'a' in e, 's' in e) for e in self.isochrons.split(',')]

        for t,hide,above,small in isochrons:
            a0 = self.cf.a0(t)
            iso = self.pf.getisochron(self.presentation, a0, self.ef)
            d10 = np.log10(iso['d'])
            self.ax.plot(d10, iso['y'], color=self.grey[0], lw=.5*self.sz_ratio)

            q = gm.where(np.abs(np.log10(iso['y']) - (self.yrange[1] - .1)) < .1)
            if not q or (d10[q[0]] < self.xrange[0]):  # (d10[q[0][0]] < self.xrange[0]):
                q= gm.where(abs(d10 - np.max([self.xrange[0] + .15, d10[0]])) < .1)

            if q and not hide:
                sx = np.mean(d10[q])
                sy = np.log10(self.pf.evaluate(self.presentation, 10**sx, a0))-.05
                y_factor = self.data_aspect
                th = np.rad2deg(np.arctan2(np.log10(self.pf.evaluate(self.presentation, 10**(sx + .3), a0))-sy, .3 * y_factor))

                self.ax.text(sx, 10 ** sy, cst.str_age(t, simple=True),
                             color=self.grey[0], size=self.scaled_pt_size*(.5 if small else .7),
                             rotation=th, rotation_mode='anchor',
                             verticalalignment='bottom' if above else 'top',
                             horizontalalignment='left',
                             bbox=dict(facecolor='none', edgecolor='none', boxstyle='square,pad=0.5'))

    def autoscale(self,xrange=None,yrange=None):
        """
        Calculate union of data ranges from all Craterplots
        Use to set default axis ranges

        xrange: optional forced range
        yrange: optional forced range

        :return: none
        """
        x0,x1,y0,y1=zip(*[cp.get_data_range(self) for cp in self.craterplot])
        mx = (.5, .9) # minimum empty margin
        my = .7 * (2 if self.presentation=='differential' else 1) # minimum empty margin
        xr = np.array([np.floor(np.log10(min(x0))-mx[0]), np.ceil(np.log10(max(x1))+mx[1])])
        yr = np.array([np.floor(np.log10(min(y0))-my), np.ceil(np.log10(max(y1))+my)])

        #set to square
        dx,dy= gm.mag(xr), gm.mag(yr)
        if self.presentation == 'differential':
            if dy/2.-np.floor(dy/2.) > .01: yr+=[0,1]
            dy= gm.mag(yr)
            d=dy/2-dx
            d0=abs(d)
            d2a=int(d0/2)
            d2b=d0-d2a
            if d<0: yr+=np.array([-d2b,d2a])*2
            if d>0: xr+=np.array([-d2a,d2b])
        else:
            d=dy-dx
            d0=abs(d)
            d2a=int(d0/2)
            d2b=d0-d2a
            if d < 0: yr += [-d2a,d2b]
            if d > 0: xr += [-d2a,d2b]

        self.xrange = xr if xrange is None else np.array(xrange,dtype=float)
        self.yrange = yr if yrange is None else np.array(yrange,dtype=float)

    def create_summary_table(self,f_out=None):
        """
        Output table of Craterplot age calculations to stdout

        :return: none
        """
        s=[]
        for cp in self.craterplot:
            if cp.type in ['c-fit', 'd-fit', 'poisson', 'b-poisson']:
                cp.calculate_age(self)
                d = {k: getattr(cp, k, None) for k in
                      {'name', 'binning', 'range', 'bin_range', 'resurf', 'type', 'source', 'n', 'n_event', 't','a0','n_d'}}
                d.update({k: getattr(cp.cratercount, k, None) for k in {'area'} })
                s+=[d]

        if not s: return

        n_d_lbl='N({:0g})'.format(self.ref_diameter)
        # w now obsolete in this table (width)
        table = (('name', '24', '', None),
                 ('area', '8', '.5g', None),
                 ('binning', '>10', '', None),
                 ('bin_range' if d['type'] in ('c-fit','d-fit') else 'range', '5', '.2g', ('d_min', 'd_max')),
                 ('type', '>9', '', ('Method',)),
                 ('resurf', '6', '', None),
                 ('n', '7', '', None),
                 ('n_event', '9', '', None),
                 ('t', '7', '.3g', ('Age','Age-','Age+')),
                 ('a0', '6', '.4g', ('a0','a0-','a0+')),
                 ('n_d', '8', '.2e', (n_d_lbl,n_d_lbl+'-',n_d_lbl+'+')),
                 ('source', '', '', None),
                 (' ', '', '', None),
                 ('Range', '', '', None),
                 ('MathML', '', '', ('Age', n_d_lbl)),
                 ('Latex', '', '', ('Age',n_d_lbl)),
                 )

        ln = []
        for k, w, f, t in table:
            if t is not None:
                ln += list(t)
            else:
                ln += [k[:1].upper() + k[1:]]
        st = ','.join(ln)

        for d in s:
            ln=[]
            for k,w,f,_ in table:
                if k=='name' and d[k]=='':
                    d[k]= gm.filename(d['source'], 'n')
                if k in ('range','bin_range','t','a0'):
                    v=','.join([('{:'+f+'}').format(e) for e in d[k]])
                    if k in ('range','bin_range'):
                        txt_range=gm.diameter_range(d[k])
                elif k=='n_d':
                    v = ','.join([('{:' + f + '}').format(10**e) for e in d['a0']])
                elif k in ('Latex','MathML'):
                    t = d['t']
                    v0 = cst.str_age(t[0], t[2] - t[0], t[0] - t[1], mu=self.mu, MathML = k=='MathML')
                    a0 = d['a0']
                    v1 = gm.scientific_notation(10**a0[0],10**a0[2],10**a0[1], unit='km-2', MathML = k=='MathML')
                    v=v0+','+v1
                elif k==' ':v=' '
                elif k == 'Range':
                    v=txt_range
                else:
                    v=('{:'+f+'}').format(d[k])
                ln+=[v]
            st += '\n'+','.join(ln)+'," "'

        st = (',,,,,,,,,,,,,,,,,,,,Formatted values [paste MathML into Word using CTRL-SHIFT-V]\n'
              ',,,,,,,,,,,,,,,,,,,,,MathML,,Latex,\n'
              )+st

        if f_out:
            try:
                gm.write_textfile(f_out, st)
            except:
                sys.exit(gm.bright("Unable to write file: ") + f_out)

        return st #for test routine





    def create_sequence_table(self,f_csv):
        """
        Output table of sequence probabilities from sequence plot

        :f_csv: filepath for csv output
        :return: sequence table
        """

        # calculate sequence table
        show_self_comparison = False
        cpl = [e for e in self.craterplot if e.type in ['poisson','buffered-poisson'] and not e.hide]
        n = len(cpl)
        if n<2: return
        s=[['' for i in range(n+1)] for j in range(n+1)]

        for i in range(n):
            cpl[i].calculate_age(self)
            s[i][0]=cpl[i].name

        for i in range(n):
            for j in range(i if show_self_comparison else i+1,n):
                P = cpl[i].pdf.calculate_sequence_probability(cpl[j].pdf)
                s[i][j+1] = gm.sigfigs(1.-P,2)
                s[j][i+1] = gm.sigfigs(P,2)

        st='Table of probability that t(x)>t(y)\n'+'\n'.join([','+','.join([cpl[i].name for i in range(n)])]+[','.join(e) for e in s])

        # calculate probability of simultaneous formation vs median time formation

        t = [cp.pdf.t(0.5) for cp in cpl]
        t_mean= sum(t)/float(n)

        st += '\nt_median:,'+','.join(['{0:.3g}'.format(tt) for tt in t])
        st += '\nmean(t_median):,'+'{0:.4g}'.format(t_mean)

        # second calculation of same

        lam_ratio = [self.cf.N1(tt)/self.cf.N1(t_mean) for tt,cp in zip(t,cpl)]
        pdf2 = [cst.Craterpdf(self.pf,self.cf,cp.cratercount,cp.range,k=cp.pdf.k,lam=cp.pdf.lam*lr) for lr,cp in zip(lam_ratio,cpl)]

        # combined pdf

        pdf3 = copy.deepcopy(pdf2[0])
        pdf3.pdf = np.product([e.pdf for e in pdf2],0)

        rel_probs =  [e.relative_probability(t0) for e,t0 in zip(pdf2,t)]
        compound_prob = np.product(rel_probs)

        worse = np.where(pdf3.pdf < compound_prob, pdf3.dt * pdf3.pdf, 0)
        prob_worse = sum(worse)/sum(pdf3.dt * pdf3.pdf)

        st += ('\n\nProbability of more distant time configuration than observed if all surfaces formed at t_mean:\n,'
               +'{0:g}'.format(prob_worse) )

        try:
            gm.write_textfile(f_csv,st)
        except:
            sys.exit(gm.bright("Unable to write file: ")+f_csv)








#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np
from matplotlib.textpath import TextPath

import craterstats as cst
import craterstats.gm as gm


class Craterplot:
    """
    Specify and draw a single overplot: either a crater data series, or an age estimate of one
    """

    def __init__(self,*args,**kwargs):
        self.UpdateSettings({
            'cratercount':None, # if not provided, will create from 'source'
            'source':'',
            'name':'',
            'range':['0','inf'],   #unconstrained
            'snap':1,
            'type':'data',
            'error_bars':1,
            'hide':0,
            'colour':0,
            'psym':0,
            'binning':'pseudo-log',
            'age_left':0,
            'display_age':1,
            'resurf':0,
            'resurf_showall':0,
            'isochron':0,
            'offset_age':[0.,0.],
            },*args,kwargs)

        self.range = self.cratercount.decode_range(self.range,self.binning,self.snap)

    def UpdateSettings(self,*args,**kwargs): #pass either dictionary or keywords
        a = {k: v for d in args for k, v in d.items()}
        a.update(**kwargs)
        for k, v in a.items():
            if k in ('offset_age'):
                v = [float(e) for e in v]
            if k == 'range':
                pass #do nothing
            if k in ('error_bars',
                     'snap',
                     'hide',
                     'age_left',
                     'display_age',
                     'resurf',
                     'resurf_showall',
                     'isochron',
                     'psym',
                     ):
                v = int(v)
            setattr(self, k, v)
        if not self.cratercount and self.source:
            self.cratercount = cst.Cratercount(self.source)
        if not self.source and self.cratercount:
            self.source = self.cratercount.filename
        if not self.name and self.source:
            self.name = gm.filename(self.source,"n")


    def calculate_age(self,cps):
        """
        Calculate age

        :param cps: Craterplotset instance
        """
        if self.type in ['poisson','b-poisson']:
            pf_range=cps.pf.range
            r0=np.clip(self.range,pf_range[0],pf_range[1])
            self.pdf=cst.Craterpdf(cps.pf, cps.cf, self.cratercount, r0, bcc=self.type == 'b-poisson')
            self.t = self.pdf.median1sigma()  # median/1-sigma gaussian-equivalent percentiles
            self.a0=cps.cf.a0(self.t)
            self.n=self.pdf.k
            self.n_event=self.n

        else:
            if self.type=='c-fit':
                p0=self.cratercount.getplotdata("cumulative",self.binning, range=self.range,
                    resurfacing=self.resurf_showall if self.resurf and self.type == 'c-fit' else None,
                    pf=cps.pf)
            elif self.type=='d-fit':
                p0=self.cratercount.getplotdata("differential",self.binning,range=self.range,pf=cps.pf)

            self.n=p0['n']    #override with number used in fit (range may be a bit different because of d_min)
            self.n_event=p0['n_event']
            self.a0=cps.pf.fit(p0)
            self.t = [cps.cf.t(a0=e) for e in self.a0]
            self.bin_range=p0['bin_range']

        self.n_d = cps.pf.evaluate("cumulative", cps.ref_diameter, self.a0[0])



    def overplot(self,cps):
        """
        Add overplot elements into figure

        :param cps: Craterplotset instance
        :return: none
        """
        if not self.cratercount or self.hide: return

        if cps.presentation=='sequence':
            self.overplot_sequence_element(cps)
            return

        p = self.cratercount.getplotdata(cps.presentation, self.binning, range=self.range,
            resurfacing=self.resurf_showall if self.resurf and self.type == 'c-fit' else None,
            pf=cps.pf)

        self.n=p['n']
        self.n_event=p['n_event']

        if self.error_bars:
            cps.ax.errorbar(np.log10(p['d']),p['y'],yerr=p['err'],fmt='none',linewidth=.7*cps.sz_ratio,ecolor=cps.grey[0])

        if self.type in ['c-fit','d-fit','poisson','b-poisson']:
            self.calculate_age(cps)

            if self.isochron:
                iso=cps.pf.getisochron(cps.presentation,self.a0[0],cps.ef)
                cps.ax.plot(np.log10(iso['d']), iso['y'], label=None,color=cps.grey[0], lw=.4*cps.sz_ratio, zorder=.9)

            expansion=np.array([.99,1.01])
            fit=cps.pf.getplotdata(cps.presentation,self.a0[0],range=self.range*expansion)
            cps.ax.plot(np.log10(fit['d']), fit['y'], label='fit', color=self.colour, lw=.7*cps.sz_ratio)

            if self.display_age:
                st=cst.str_age(self.t[0], self.t[2] - self.t[0], self.t[0] - self.t[1], cps.sig_figs, mu=cps.mu)
                xy = cps.data_to_axis((np.log10(fit['d'][0]),fit['y'][0]))
                x,y = xy + 0.02*np.ones(2)*(-1 if self.age_left else 1) +np.array(self.offset_age)/(cps.decades[0]*20) #(cps.decades[0]*10).
                cps.ax.text(x,y,st,transform=cps.ax.transAxes,
                            color=self.colour,size=cps.scaled_pt_size*1.2,
                            horizontalalignment='right' if self.age_left else 'left',)

                if self.type in ['poisson','b-poisson']:
                    text_extent = TextPath((0, 0), st, size=cps.scaled_pt_size*1.2).get_extents()
                    h,w=text_extent.height,text_extent.width
                    f = 1 / (cps.cm2inch * (cps.position[2] - cps.position[0]) * 100) #conversion for axes coord
                    offset = self.pdf.offset(self.age_left)          # normalised units of mini-plot width in +x direction
                    box = np.array([.12, .05])*cps.sz_ratio          # dimensions of plot box

                    if self.age_left:  # offset from string write position
                        dx = -(f * w + .03) + (-1 + offset) * box[0]
                    else:
                        dx = f * w + .03 + offset * box[0]
                    dy = f * h / 2

                    pos = np.array([x+dx, y-dy, x+dx + box[0], y-dy + box[1]])
                    pos2=cps.axis_to_fig(pos)
                    pos3=np.concatenate([pos2[0:2],pos2[2:4]-pos2[0:2]])
                    ax = cps.fig.add_axes(pos3)
                    self.pdf.plot(ax,pt_size=cps.scaled_pt_size,color=self.colour)

        legend_label=self.make_legend_label(cps)
        cps.ax.plot(np.log10(p['d']),p['y'],label=', '.join(legend_label) if legend_label else None,
                    **cps.marker_def[self.psym],ls='',color=self.colour)



    def make_legend_label(self,cps):
        legend_label = []

        if self.type in ['c-fit', 'd-fit', 'poisson', 'b-poisson']:
            if 'c' in cps.legend:
                if self.cratercount.buffered:
                    legend_label += ['{:.1f}'.format(self.n_event)]
                else:
                    if np.abs(self.n_event - self.n) < .001:
                        legend_label += ['{:0g}'.format(self.n)]
                    else:
                        legend_label += ['{0:.1f} (of {1:d})'.format(self.n,self.n_event)]
                legend_label[-1] += " craters"
            if 'r' in cps.legend:
                if not self.cratercount.prebinned and self.type in ['poisson','b-poisson']:
                    r=self.range
                else:
                    r=gm.range(self.cratercount.generate_bins(self.binning,self.range,expand=False))
                legend_label += [gm.diameter_range(r)]
            if 'N' in cps.legend:
                legend_label += ['N({0:0g})'.format(cps.ref_diameter) +'=' + gm.scientific_notation(10**self.a0[0],10**self.a0[2],10**self.a0[1], unit='km-2')]
            if cps.presentation == 'sequence':
                if 'a' in cps.legend:
                    legend_label+=[gm.scientific_notation(self.cratercount.area, sf=3, unit='km2')]
                if 'A' in cps.legend:
                    legend_label+=[(cst.str_age(self.t[0], self.t[2] - self.t[0], self.t[0] - self.t[1], cps.sig_figs,
                                            mu=cps.mu, no_error=False))]

        if self.type=='data':
            if 'n' in cps.legend:
                legend_label+=[self.name if self.name!='' else gm.filename(self.source, "n")]
            if 'a' in cps.legend:
                legend_label+=[gm.scientific_notation(self.cratercount.area, sf=3, unit='km2')]
            if 'p' in cps.legend:
                if self.cratercount.perimeter:
                    legend_label += [gm.scientific_notation(self.cratercount.perimeter, sf=3) + ' km']

        return legend_label


    def get_data_range(self,cps):
        """
        Return data range of Craterplot

        :param cps: Craterplotset instance
        :return: (d_min,d_max,y_min,y_max)
        """
        p = self.cratercount.getplotdata(cps.presentation, self.binning, range=self.range,
             resurfacing=self.resurf_showall if self.resurf and self.type == 'c-fit' else None,
             pf=cps.pf)
        return gm.range(p['d']) + gm.range(p['y'])




    def overplot_sequence_element(self,cps):
        """
        Add overplot sequence elements into figure

        :param cps: Craterplotset instance
        :return: none
        """

        p = self.cratercount.getplotdata('differential', self.binning, range=self.range,
            resurfacing=None, pf=cps.pf)

        self.n=p['n']
        self.n_event=p['n_event']

        if self.type in ['c-fit','d-fit','poisson','b-poisson']:

            self.calculate_age(cps)

            seq=[e for e in cps.craterplot if e.type != 'data' and not e.hide]
            n=len(seq)
            i=[i for i,e in enumerate(seq) if e is self][0]
            dy = (n - i) / (n+1)
            w0 = 0.5/(n+1)
            w = 0.3 * w0

            lw = cps.scaled_pt_size / 16.

            for j in [0, 1] if cps.crossover else [0]:

                if self.type in ['poisson','b-poisson']:
                    self.pdf.violin_plot(cps.ax2[j],dy,w0,pt_size=cps.scaled_pt_size,color=self.colour,invert=cps.invert)

                if self.type in ['d-fit','c-fit']:
                    cps.ax2[j].fill_between([self.t[1], self.t[2]], [dy + w, dy + w], [dy - w, dy - w],
                                            color=gm.mix_colours(self.colour,'black' if cps.invert else 'white',0.35),
                                            edgecolor=self.colour,lw=lw,zorder=4)

                    cps.ax2[j].plot([self.t[0], self.t[0]],
                                    [dy + w, dy - w], color=self.colour, alpha=0.5, lw=lw*.7,zorder=4)

            text = '\n'.join(gm.strip_quotes(self.name).split(r'\n')) #clear quotes and substitute real linebreaks
            cps.ax.text(-.007, dy, text, fontsize=cps.scaled_pt_size * .75, ha='right', va='center')

            lbl = self.make_legend_label(cps)
            if lbl: cps.ax.text(1.007, dy, '\n'.join(lbl), fontsize=cps.scaled_pt_size * .75, ha='left', va='center')

















#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

def write_textfile(filename,s):
    '''
    write string or list of strings to text file

    :param filename: filename
    :param s: string or list of strings
    :return: none
    '''
    
    with open(filename, 'w') as file:
        file.writelines('\n'.join(s) if type(s) is list else s)


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import unittest

import craterstats.cli as cli

class Testcli(unittest.TestCase):

    def test_decode_abbreviation(self):
        s=['Red','Green','Blue','Moon, Neukum (1983)','Mars, Hartmann & Daubar (2016)']
        # self.assertEqual(cli.decode_abbreviation(s, '2'), 2)
        # self.assertEqual(cli.decode_abbreviation(s, '2', one_based=True), 1)

        self.assertEqual(cli.decode_abbreviation(s, 'gr'), 1)
        self.assertEqual(cli.decode_abbreviation(s, 'gn'), 1)
        self.assertEqual(cli.decode_abbreviation(s, 'neukum83'), 3)
        self.assertEqual(cli.decode_abbreviation(s, 'neukum83',one_based = True), 3)
        self.assertEqual(cli.decode_abbreviation(s, 'h&d'), 4)

        self.assertRaises(SystemExit,cli.decode_abbreviation, s, 'error')
        self.assertEqual(cli.decode_abbreviation(s, 'r', allow_ambiguous=True), 0)
        self.assertRaises(SystemExit, cli.decode_abbreviation, s, 'r')


if __name__ == '__main__':
    unittest.main()



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from .idl import *
from .maths import *
from .file import *
from .plotting import *
from .string import *



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np
import craterstats.gm as gm

def diameter_range(d,sf=2):
    """
    Convert numerical diameters in km into string diameter range with units

    :param d: two element numerical sequence of diameters in km
    :param sf: no of significant figures
    :return: string diameter range
    """

    # \u2013 en-dash \u2009 short space
    if np.isinf(d[1]):
        return 'unconstrained'

    if d[1] >=1:
        s = gm.sigfigs(d, sf)
        return '{0[0]}\u2013{0[1]} km'.format(s)
    else:
        s = gm.sigfigs(np.array(d)*1000, sf)
        return '{0[0]}\u2013{0[1]} m'.format(s)


if __name__ == '__main__':
    print(diameter_range(np.array([.65, 1.2])))
    print(diameter_range([.65, 1.2]))
    print(diameter_range([.65, .95]))


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import unittest

import numpy as np

import craterstats as cst

class TestCraterstats(unittest.TestCase):

    def test_bin_bias_correction(self):
        self.assertEqual(cst.bin_bias_correction(np.sqrt(2), 0), 1.)
        self.assertAlmostEqual(cst.bin_bias_correction(np.sqrt(2), -2.), 1.015,places=3)

    def test_str_age(self):
        self.assertEqual(cst.str_age(.314, .11, .14),'$314^{+110}_{-140}$ Ma')
        self.assertEqual(cst.str_age(.314, .11, .14, unit='Ga', mu=True), '$\mu0.314^{+0.11}_{-0.14}$ Ga')
        self.assertEqual(cst.str_age(.314, simple=True), '314 Ma')
        self.assertEqual(cst.str_age(1.), '$1.00$ Ga')
        self.assertEqual(cst.str_age(1., simple=True), '1 Ga')
        self.assertEqual(cst.str_age(.314, .11, .14,sf=2),'$310^{+100}_{-100}$ Ma')

if __name__ == '__main__':
    unittest.main()



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np
import re

import craterstats as cst
import craterstats.gm as gm


class Epochs:
    """
    Decode Epoch definitions, calculate derived values, and provide overplot routines

    """

    def __init__(self, source,identifier,pf,cf):

        if type(source) is dict:
            src = source
        else:
            if '\n' in source: # multiline string is definition
                txt = source + '\nepochs={\n name="null"\n}' # add null entry to force implied array
            else: # single line string is filename
                txt = gm.read_textfile(source, as_string=True, ignore_hash=True)
            src = gm.read_textstructure(txt, from_string=True)

        definition = next((e for e in src['epochs'] if e['name'] == identifier), None)
        if definition is None:
            raise ValueError('Epoch system not found: ' + identifier)

        self.name = definition['name']
        self.epoch = definition['epoch']
        self.formatting = definition['formatting']
        self.ref_diameter = [int(e) for e in definition['ref_diameter']] if 'ref_diameter' in definition else None
        self.density = [float(e) for e in definition['density']] if 'density' in definition else None
        self.time = [float(e) for e in definition['time']] if 'time' in definition else None
        self.reference = definition['reference']
        self.pf=pf
        self.cf=cf

        if self.time is None:
            a0=[pf.fit({'presentation':'cumulative','d':d,'y':ncum/1e6,'err':0})[0] for d,ncum in zip(self.ref_diameter,self.density)]
            self.time=[0.]+list(cf.t(a0=np.array(a0)))

    def __str__(self):
        return self.name

    def decode_formatting(self):
        """
        Decode epoch definition

        """
        label_slant=int(self.formatting[0])
        colour = [int(re.search('\d?', e)[0]) for e in self.formatting[1:]]
        offset = ['r' in e for e in self.formatting[1:]]
        boundary = ['b' in e for e in self.formatting[1:]]
        return label_slant,colour,offset,boundary

    def chronology_oplot(self,cps,phi=False):
        """
        Overplot epoch system on chronology/rate plot

        :param cps: Craterplotset instance
        :param phi: rate function?
        :return: none
        """
        p_t,p_y=self.cf.getplotdata(phi=phi,linear=True)
        label_slant, colour, offset, boundary = self.decode_formatting()

        for i,(t0,t1) in enumerate(zip(self.time[1:],self.time[2:])):  # fill epoch regions
            t=np.linspace(t0,t1,100)
            y=np.interp(t,p_t,p_y)
            cps.ax.fill_between(t,y,y*1e-11, color=cps.grey[3-colour[i+1]], edgecolor=None, lw=0)

        for i, t in enumerate(self.time[1:]):  # add dividing lines
            y = np.interp(t, p_t, p_y)
            cps.ax.plot([t,t], [y, y * 1e-11], color=cps.grey[0] ,lw=.4 if boundary[i] else .3 ,
                        alpha=1 if boundary[i] else .3)

        t0 = list(self.time) + [cps.xrange[0]]
        for i,e in enumerate(self.epoch): # write epoch names
            t=(t0[i]+t0[i+1])/2
            y = self.cf.phi(t) if phi else self.cf.N1(t)
            cps.ax.text(t-.03, y*1.5, e,
                        size=cps.pt_size * .7, rotation=label_slant,
                        rotation_mode='anchor', verticalalignment='center', horizontalalignment='left',
                        bbox=dict(facecolor='none', edgecolor='none', pad=50))

        y = self.cf.phi(self.time[1])*.7 if phi else self.cf.N1(self.time[1])/2.5
        for i,t in enumerate(self.time): # write epoch times
            if t!=0:
                cps.ax.annotate(' {:0.3g} '.format(t),xy=(t, y),
                            xytext=(cps.pt_size * .04, 0), textcoords="offset points",
                            size=cps.pt_size * .7, rotation=90,
                            verticalalignment='bottom' if offset[i] else 'top',
                            horizontalalignment='center',
                            bbox=dict(facecolor='none', edgecolor='none', pad=50))


    def sequence_oplot(self,cps):
        """
        Overplot epoch system on a sequence plot

        :param cps: Craterplotset instance
        :return: none
        """

        label_slant, colour, offset, boundary = self.decode_formatting()

        for j in [0, 1] if cps.crossover else [0]:

            for i,(t0,t1) in enumerate(zip(self.time[1:],self.time[2:])):  # fill epoch regions
                cps.ax2[j].fill_between([t0,t1],[0,0],[1,1], color=cps.grey[3-colour[i+1]], edgecolor=None, lw=0)

            for i, t in enumerate(self.time[1:]):  # add dividing lines
                cps.ax2[j].plot([t,t], [0,1], color=gm.mix_colours(cps.grey[0], 'black' if cps.invert else 'white' ,1 if boundary[i] else .3),
                                lw=.4 if boundary[i] else .3)

            t0 = sorted(self.time + [cps.t_min,cps.t_max])[1:] #lower limit substitutes correct place in seq

            for i,e in enumerate(self.epoch): # write epoch names
                lbl=[a[0] for a in re.split(' |-',e)]
                lbl=lbl[0].lower()+lbl[1] if len(lbl)==2 else lbl[0]
                if j==0:
                    t = (t0[i] + t0[i + 1]) / 2
                else:
                    t = np.sqrt(t0[i] * t0[i + 1])
                cps.ax2[j].text(t, .01, lbl, horizontalalignment='center',
                                size=cps.scaled_pt_size * .55, clip_on=True)



    def oplot(self,cps):
        """
        Overplot epoch system on crater count plot

        :param cps: Craterplotset instance
        :return: none
        """
        a0=[self.cf.a0(t) for t in self.time[1:]]
        iso = [self.pf.getisochron(cps.presentation,e,cps.ef) for e in a0]
        _, colour, _, boundary = self.decode_formatting()

        for i,(a,b) in enumerate(zip(iso,iso[1:])):
            x=np.concatenate((a['d'], b['d'][::-1]))
            y = np.concatenate((a['y'], b['y'][::-1]))
            cps.ax.fill(np.log10(x),y, color=cps.grey[3 - colour[i + 1]], edgecolor=None, lw=0)

            if boundary[i]:
                cps.ax.plot(np.log10(a['d']),a['y'], color=cps.grey[0], lw=.3)



if __name__ == '__main__':
    f="config/functions.txt"
    cf = cst.Chronologyfn(f, 'Moon, Neukum (1983)')
    pf = cst.Productionfn(f, 'Moon, Neukum (1983)')
    ep=Epochs(f,'Moon, Wilhelms (1987)',pf,cf)



















#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import unittest
from unittest.mock import patch
import textwrap

import io
import numpy as np

import craterstats as cst
import craterstats.gm as gm

class TestPlottingClasses(unittest.TestCase):

    root = cst.gm.filename(cst.__file__, 'p')
    file_fns = root + 'config/functions.txt'
    cf = cst.Chronologyfn(file_fns, 'Mars, Hartmann & Neukum (2001)')
    pf = cst.Productionfn(file_fns, 'Mars, Ivanov (2001)')
    cc = cst.Cratercount(root + 'sample/Pickering.scc')

    def test_Craterplotset_CreatePlotSpace(self):
        cps=cst.Craterplotset()
        cps.CreatePlotSpace()
        self.assertTrue(cps.fig)
        self.assertTrue(cps.ax)

    def test_Craterplot_get_data_range(self):
        cp = cst.Craterplot(cratercount=self.cc)
        cps = cst.Craterplotset(craterplot=[cp])

        res = cp.get_data_range(cps)
        self.assertTrue(np.allclose(res,(.1,2.7,6.5e-4,8.2),rtol=.05))

    def test_Craterplotset_autoscale(self):
        cp = cst.Craterplot(cratercount=self.cc)
        cps = cst.Craterplotset(craterplot=[cp])
        cps.autoscale()
        self.assertEqual(list(cps.xrange)+list(cps.yrange),[-2, 2, -5, 3])

    def test_Craterplotset_summary(self):
        cp = cst.Craterplot(cratercount=self.cc,range=[.2,.7],type='d-fit')
        cps = cst.Craterplotset(cf=self.cf,pf=self.pf,craterplot=[cp])

        def get_cps_summary():
            res0=cps.create_summary_table()
            res=res0.split('\n')
            res=','.join(res[3].split(',')[0:16]) # trim off headers/source file
            return res

        # the following results are not fundamental, but verified against CraterstatsII (see demo plots):
        # name,area,binning,d_min,d_max,method,resurf,n,n_event,age,age-,age+,a0,a0-,a0+,N(1)
        self.assertEqual(get_cps_summary(),
            'Pickering,3036.6,pseudo-log,0.2,0.7,d-fit,0,313.0,313,0.668,0.613,0.722,-3.488,-3.524,-3.454,3.25e-04')
        cp.UpdateSettings(binning='10/decade')
        self.assertEqual(get_cps_summary(),
            'Pickering,3036.6,10/decade,0.2,0.79,d-fit,0,313.0,313,0.645,0.593,0.696,-3.503,-3.539,-3.469,3.14e-04')
        cp.UpdateSettings(type='c-fit', resurf=1, binning='pseudo-log')
        self.assertEqual(get_cps_summary(),
            'Pickering,3036.6,pseudo-log,0.2,0.7,c-fit,1,313.0,313,0.691,0.654,0.729,-3.472,-3.497,-3.449,3.37e-04')
        cp.UpdateSettings(type='poisson', range=[.22,.43], resurf=0)
        self.assertEqual(get_cps_summary(),
            'Pickering,3036.6,pseudo-log,0.22,0.43,poisson,0,223,223,0.685,0.641,0.733,-3.476,-3.505,-3.447,3.34e-04')
        cp.UpdateSettings(type='poisson', range=[.22, .43], binning='none')
        self.assertEqual(get_cps_summary(),
            'Pickering,3036.6,none,0.22,0.43,poisson,0,223,223,0.685,0.641,0.733,-3.476,-3.505,-3.447,3.34e-04')







#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np
import re

def sigfigs(n, digits):

    if isinstance(n, (list, tuple, np.ndarray)):
        return [sigfigs(e,digits) for e in n]

    f="{:."+str(digits-1)+"e}"
    s=f.format(n)
    
    m=re.search("(?P<sign>-?)(?P<num>[\d\.]*)e(?P<exp>.*)",s)    
    sf=m['num'].replace('.','')    
    e=int(m['exp'])

    #construct output for 4 different cases   
    if e<0:                             #insert leading zeros
        res='0.'+'0'*(abs(e)-1)+sf  
    elif e<digits-1:                    #insert intermediate decimal point
        res=sf[:e+1]+'.'+sf[e+1:]
    elif e<digits-1:                    #no decimal required
        res=sf
    else:                               #add trailing zeros
        res=sf+'0'*(e-digits+1)
        
    res=m['sign']+res
    
    return res

if __name__ == '__main__':
    print(sigfigs(3.14159,3))
    print(sigfigs(3.14159,2))
    print(sigfigs(112.14159,2)) # should be '110'
    print(sigfigs(-.0314159,3))
    print(sigfigs(314759,3))
    print(sigfigs(31.4759,1))
    print(sigfigs([243.3,0.3422],2))


#  Copyright (c) 2024, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import matplotlib.colors as mc

def mix_colours(c1, c2, amount=0.5):
    '''
    matplotlib: mix two colours to proportion (sometimes preferable to alpha over background)

    :param c1: first colour
    :param c2: second colour
    :param amount: fraction of first
    :return: name of first available
    '''
    c = tuple(map(lambda x, y: y * (1. - amount) + x * amount, mc.to_rgb(c1), mc.to_rgb(c2)))
    return c


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import colorama

def bright(txt):
    '''
    Wrap string with control codes to cause bright display on console
    '''

    if "init" not in bright.__dict__:
        colorama.init()
        bright.init = True
    return colorama.Style.BRIGHT+txt+colorama.Style.RESET_ALL

if __name__ == '__main__':
    print(bright("bright") + " normal")
    print(bright.init)
    print(bright("bright") + " normal")


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import unittest
import textwrap

import numpy as np

import craterstats as cst

class TestCraterFunctionObjects(unittest.TestCase):

    file_fns = cst.gm.filename(cst.__file__, 'p') + 'config/functions.txt'

    def test_Chronologyfn(self):
        cf = cst.Chronologyfn(self.file_fns, 'Moon, Neukum (1983)')
        self.assertEqual(cf.N1(0.),0.)
        N1=5.44E-14*(np.exp(6.93*3.)-1)+8.38E-4*3.
        self.assertEqual(cf.N1(3.), N1)
        self.assertEqual(cf.a0(3.), np.log10(N1))
        self.assertAlmostEqual(cf.phi(0.), 8.38E-4)
        self.assertAlmostEqual(cf.t(n1=N1), 3.,places=4)

        #user defined function
        f = textwrap.dedent("""
            chronology={
              name='user_function'
              n1_code="
                a=5.44E-14 * (exp(6.93*t)-1)
                b=8.38E-4 * t
                n1=a+b
                "
            }
            """)
        cf_user = cst.Chronologyfn(f, 'user_function')
        self.assertEqual(cf_user.N1(3.), N1)
        self.assertTrue(np.allclose(cf.getplotdata(), cf_user.getplotdata()))
        self.assertTrue(np.allclose(cf.getplotdata(phi=True),cf_user.getplotdata(phi=True))) # user fn same as pre-defined

    def test_Productionfn(self):
        Npf = cst.Productionfn(self.file_fns,'Moon, Neukum (1983)') # polynomial PF
        self.assertTrue(np.allclose(Npf.range,(.01, 300.)))
        self.assertTrue(np.allclose(Npf.a, [-2.5339, -3.6269, 0.43662, 0.79347, 0.086468, -0.26485, -0.066382, 0.037923,
                                           0.010596, -0.0022496, -0.00051797, 0.0000397]))
        Hpf = cst.Productionfn(self.file_fns, 'Mars, Hartmann (2005)')  # tabular PF
        test_a0=1.5
        for pf in [Npf,Hpf]:
            for presentation in ['cumulative','differential']:
                r=pf.getplotdata(presentation,a0=test_a0) # create isochron with given a0
                r['err']=np.ones(len(r['y']))
                self.assertAlmostEqual(pf.fit(r)[0],test_a0) # test whether fitting can recover a0

    def test_Equilibriumfn(self):
        ef = cst.Productionfn(self.file_fns,'Standard lunar equilibrium (Trask, 1966)',equilibrium=True)
        self.assertEqual(ef.evaluate('cumulative', 1., a0=ef.a[0]), 10 ** (-1.1 - 2. * np.log10(1.)))
        self.assertEqual(ef.evaluate('cumulative', 10., a0=ef.a[0]), 10 ** (-1.1 - 2. * np.log10(10.)))

    def test_Epochs(self):
        cf = cst.Chronologyfn(self.file_fns, 'Mars, Hartmann & Neukum (2001)')
        pf = cst.Productionfn(self.file_fns, 'Mars, Ivanov (2001)')
        ep = cst.Epochs(self.file_fns, 'Mars, Michael (2013)', pf, cf)
        self.assertTrue(np.allclose(ep.time,[0.0, 0.328, 1.23, 3.37, 3.61, 3.71, 3.83, 3.94],rtol=.01)) # (Michael, 2013)



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

with open(here / "README.md", encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(here / 'requirements.txt') as f:
    required = f.read().splitlines()

setup(
    author="Greg Michael",
    author_email="ggmichael@users.noreply.github.com",
    python_requires=">=3.8, <4",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Craterstats - a tool to analyse and plot crater count data for planetary surface dating",
    entry_points={
        "console_scripts": [
            'craterstats=craterstats.cli:main',
        ],
    },
    license="BSD license",
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords=["craters"],
    name="craterstats",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    package_data={
        "craterstats": ["config/*.txt", "sample/*.diam", "sample/*.r", "sample/*.scc",  "sample/*.binned", "sample/*.stat", "sample/README.md"]
    },
    test_suite="tests",
    tests_require=required,
    install_requires=required,
    url="https://github.com/ggmichael/craterstats",
    version="3.1.1",
)



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

from ..maths.range import range as gm_range
from ..maths.range import mag as gm_mag


def ticks(x, max_ticks, max_minor=None):
    '''
    calculate nice number tick values for plot axis

    :param x: plot data vector
    :param max_ticks: max no of major ticks
    :param max_minor: max no of minor ticks (in total, i.e. n_ticks*minor_ticks) (optional)
    :return: tick values
    '''


    if not max_ticks: max_ticks = 5
    max_ticks = np.clip(max_ticks, 2, None)

    range = gm_range(x)
    mag = gm_mag(range)

    multiplier = 1.

    while True:
        estimate = mag / float(max_ticks * multiplier)
        exponent = np.floor(np.log10(estimate))
        fraction = estimate / 10. ** exponent
        a = np.searchsorted([0, 1, 2, 5], fraction)-1
        nice_fraction = ([1, 2, 5, 10])[a]
        step = nice_fraction * 10. ** exponent
        i = [np.floor(range[0] / step), np.ceil(range[1] / step)]
        ticks = (i[0] + np.arange(gm_mag(i) + 1)) * step
        n_ticks = len(ticks)
        multiplier *= .9
        if n_ticks <= max_ticks:
            break

    if max_minor:
        choice = [1, 2, 4, 10] if nice_fraction==2 else [1,2,5,10]
        b=max_minor/(n_ticks-1)
        j = np.searchsorted(choice,b)
        minor_ticks=choice[j]
        return ticks,minor_ticks

    return ticks



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from .ticks import ticks
from .mpl_check_font import mpl_check_font
from .mix_colours import mix_colours



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from .bright import bright
from .scientific_notation import scientific_notation
from .sigfigs import sigfigs
from .strip_quotes import strip_quotes
from .quoted_split import quoted_split
from .diameter_range import diameter_range



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import craterstats.gm as gm

def str_age(age,e1=0,e2=0,sf=3,unit=None,simple=False,no_error=False,mu=False,MathML=False):
    '''
    Convert numerical age in Ga to latex-style string, optionally with sub/superscript errors and mu uncertainty function

    :param age:
    :param e1: plus error
    :param e2: minus error
    :param sf: no of sig figs
    :param unit: force specified unit
    :param simple: don't show zero decimals (for isochrons), nor use latex-style
    :param no_error: don't show errors
    :param mu: show mu uncertainty function
    :param MathML - use MathML instead of latex. Can paste into MS Word
    :return: age string
    '''

    f = [1, 1e-3, 1e-6, 1e-9]
    unit0 = ['Ga', 'Ma', 'ka', 'a']

    if unit is not None:
        a = next((e, u) for e, u in zip(f, unit0) if unit == u)
    else:
        a = next((e, u) for e, u in zip(f, unit0) if age >= e)

    v=age/a[0]

    if simple:
        st='{:g}'.format(v)
    else:
        st= gm.sigfigs(v, sf)

    if e1>0 and not no_error :
        e1s = gm.sigfigs(e1 / a[0], sf - 1)
        e2s = gm.sigfigs(e2 / a[0], sf - 1)
        if MathML:
            st = '<msubsup><mtext>'+st+'</mtext><mtext>&#x2212;'+e2s+'</mtext><mtext>+'+e1s+'</mtext></msubsup>'
        else:
            st += '^{+' + e1s + '}_{-' + e2s + '}'

    if simple:
        st += ' ' + a[1]
    elif MathML:
        st = ('<math xmlns="http://www.w3.org/1998/Math/MathML">'
              +('<mi>&#956;</mi>' if mu else '')+st+'<mtext>&#8239;'+a[1]+'</mtext></math>')
    else:
        st = '$' + ('\mu' if mu else '') + st + '$ ' + a[1]

    return st


if __name__ == '__main__':
    print(str_age(.314,.11,.14))
    print(str_age(.314, .11, .14,unit='Ga',mu=True))
    print(str_age(.314, simple=True))
    print(str_age(.314, .11, .14, unit="ka"))
